#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
tools related to sampling
"""
# These functions are consolidated and adapted from samples.py and cut.py in
# branches/UQ/math/ for general use.
# sample() is from cut.py, and samplepts() is from examples2/seesaw2d.py, and
# everything else is from samples.py

# SAMPLING #
def _bounded_samples(lb, ub, samples, dist=None, clip=False):
  """
ensure randomness from distribution yields samples bounded by given lb & ub

Inputs:
    lb -- a list of the lower bounds (e.g. lb = [0,3])
    ub -- a list of the upper bounds (e.g. ub = [2,4])
    samples -- a list of sample points (e.g. [[1,3],[0,4],[2,3],[2,4]])
    dist -- a mystic.tools.Distribution instance (or list of Distributions)
    clip -- if True, clip at bounds, else resample [default = False]

Notes: if clip is None, do not clip or resample (may exceed bounds)
"""
  import numpy as np
  samples = np.asarray(samples).T
  if dist is None:
    return samples
  npts = samples.T.shape[0]
  if not npts: return samples
  if hasattr(dist, '__len__'): #FIXME: isiterable
    pts = np.array(tuple(di(npts) for di in dist)).T + samples.T
  else:
    pts = dist((npts,len(lb))) + samples.T # transpose of desired shape
    dist = (dist,)*len(lb)
  if clip is None: return pts.T
  bad = ((pts == lb) + (pts == ub)).T #NOTE: these points are ok
  pts = np.clip(pts, lb, ub).T
  if clip: return pts  #XXX: returns a numpy.array
  bad = ((pts.T == lb) + (pts.T == ub)).T != bad
  new = bad.sum(-1)
  _n, n = 1, 1000
  while any(new):
    if _n == n: #XXX: slows the while loop...
      raise RuntimeError('bounds could not be applied in %s iterations' % n)
    for i,inew in enumerate(new): #XXX: slows... but enables iterable dist
      if inew: pts[i][bad[i]] = dist[i](inew) + samples[i][bad[i]]
    bad = ((pts.T == lb) + (pts.T == ub)).T #NOTE: these points are ok
    pts = np.clip(pts.T, lb, ub).T
    bad = ((pts.T == lb) + (pts.T == ub)).T != bad
    new = bad.sum(-1)
    _n += 1
  return pts


def _random_samples(lb, ub, npts=10000):
  """
generate npts random samples between given lb & ub

Inputs:
    lb -- a list of the lower bounds
    ub -- a list of the upper bounds
    npts -- number of sample points [default = 10000]
"""
  from mystic.tools import random_state
  dim = len(lb)
  pts = random_state(module='numpy.random').rand(dim,npts)
  for i in range(dim): #XXX: use array operations?
    lbi = lb[i]
    ubi = ub[i]
    pts[i] = (pts[i] * abs(ubi - lbi)) + lbi
  return pts  #XXX: returns a numpy.array
 #return [list(i) for i in pts]


def random_samples(lb, ub, npts=10000, dist=None, clip=False):
  """
generate npts samples from the given distribution between given lb & ub

Inputs:
    lb -- a list of the lower bounds
    ub -- a list of the upper bounds
    npts -- number of sample points [default = 10000]
    dist -- a mystic.tools.Distribution instance (or list of Distributions)
    clip -- if True, clip at bounds, else resample [default = False]

Notes: if clip is None, do not clip or resample (may exceed bounds)
"""
  if dist is None:
    return _random_samples(lb,ub, npts)
  import numpy as np
  pts = np.zeros((npts,len(lb)))
  return _bounded_samples(lb, ub, pts, dist=dist, clip=clip)


def sample(f, lb, ub, npts=10000, map=None):
  """
return number of failures and successes for some boolean function f

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, atleast_2d
  pts = _random_samples(lb, ub, npts)

  results = list(map(f, atleast_2d(transpose(pts)).tolist()))
  failure = results.count(False)
  success = len(results) - failure
  return failure, success


# STATISTICS #
def sampled_mean(f, lb, ub, npts=10000, map=None):
  """
use random sampling to calculate the mean of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  pts = _random_samples(lb, ub, npts)
  return _expectation_given_samples(f, pts, map)

# if map is None:
#   from builtins import map
# from numpy import inf, transpose, atleast_2d
# from mystic.tools import wrap_bounds
# pts = _random_samples(lb, ub, npts)
# f = wrap_bounds(f,lb,ub)

# if len(lb) != 1:
#   results = map(f, atleast_2d(transpose(pts)).tolist())
# else: #FIXME: fails for len(lb) == 1 (and the above works!)
#   _f = lambda x: f(float(x))
#   results = map(_f, transpose(pts))
# #ave = 0; count = 0
# #for Fx in results:
# #  if Fx != inf: # outside of bounds evaluates to inf
# #    ave += Fx
# #    count += 1
# results = list(filter(lambda Fx: Fx != inf, results))
# count = len(results)
# if not count: return None  #XXX: define 0/0 = None
# ave = sum(results)
# ave = float(ave) / float(count)
# return ave


def sampled_variance(f, lb, ub, npts=10000, map=None): #XXX: could be improved
  """
use random sampling to calculate the variance of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  pts = _random_samples(lb, ub, npts)
  return _variance_given_samples(f, pts, map)


def sampled_pof(f, lb, ub, npts=10000, map=None):
  """
use random sampling to calculate probability of failure for a function

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  pts = _random_samples(lb, ub, npts)
  return _pof_given_samples(f, pts, map)


def sampled_ptp(f, lb, ub, npts=10000, map=None):
  """
use random sampling to calculate the spread of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  pts = _random_samples(lb, ub, npts)
  return _ptp_given_samples(f, pts, map)


def sampled_minimum(f, lb, ub, npts=10000, map=None):
  """
use random sampling to calculate the minimum of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  pts = _random_samples(lb, ub, npts)
  return _minimum_given_samples(f, pts, map)


def sampled_maximum(f, lb, ub, npts=10000, map=None):
  """
use random sampling to calculate the maximum of a function

Inputs:
    f -- a function that takes a list and returns a number
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    npts -- the number of points to sample [Default is npts=10000]
    map -- the mapping function [Default is builtins.map]
"""
  pts = _random_samples(lb, ub, npts)
  return _maximum_given_samples(f, pts, map)


# ALTERNATE: GIVEN SAMPLE POINTS #
def _pof_given_samples(f, pts, map=None):
  """
use given sample pts to calculate probability of failure for function f

Inputs:
    f -- a function that returns True for 'success' and False for 'failure'
    pts -- a list of sample points
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, atleast_2d
  results = list(map(f, atleast_2d(transpose(pts)).tolist()))
  pof = float(results.count(False)) / float(len(results))
  return pof


def _minimum_given_samples(f, pts, map=None):
  """
use given sample pts to calculate minimum for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, atleast_2d
  return min(list(map(f, atleast_2d(transpose(pts)).tolist())))


def _expectation_given_samples(f, pts, map=None):
  """
use given sample pts to calculate expected value for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, mean, atleast_2d
  return mean(list(map(f, atleast_2d(transpose(pts)).tolist())))


def _variance_given_samples(f, pts, map=None):
  """
use given sample pts to calculate expected variance for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, var, atleast_2d
  return var(list(map(f, atleast_2d(transpose(pts)).tolist())))


def _maximum_given_samples(f, pts, map=None):
  """
use given sample pts to calculate maximum for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, atleast_2d
  return max(list(map(f, atleast_2d(transpose(pts)).tolist())))


def _ptp_given_samples(f, pts, map=None):
  """
use given sample pts to calculate spread for function f

Inputs:
    f -- a function that returns a single value, given a list of inputs
    pts -- a list of sample points
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import transpose, ptp, atleast_2d
  return ptp(list(map(f, atleast_2d(transpose(pts)).tolist())))


def sampled_pts(pts, lb, ub, map=None):
  """
determine the number of sample points inside the given bounds

Inputs:
    pts -- a list of sample points
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    map -- the mapping function [Default is builtins.map]
"""
  if map is None:
    from builtins import map
  from numpy import inf, transpose, atleast_2d
  from mystic.tools import wrap_bounds
  def zero(x):
    return 0
  f = wrap_bounds(zero,lb,ub)
  results = map(f, atleast_2d(transpose(pts)).tolist())
  #npts = len(list(filter(lambda Fx: Fx != inf, results)))
  npts = list(results).count(0) # outside of bounds evaluates to inf
  return npts

def sampled_prob(pts, lb, ub, map=None):
  """
calculates probability by sampling if points are inside the given bounds

Inputs:
    pts -- a list of sample points
    lb -- a list of lower bounds
    ub -- a list of upper bounds
    map -- the mapping function [Default is builtins.map]
"""
  prob = float(sampled_pts(pts,lb,ub,map)) / float(len(pts[0]))
  return prob






#def minF(x):
#  return model(x)

#def maxF(x):
#  return -model(x)

def alpha(n, diameter, epsilon=0.01):
  from math import log
 #return diameter * n**(-0.5) * (-log(epsilon))**(0.5)
  return diameter * (-log(epsilon) / (2.0 * n))**(0.5)


#######################################################################


if __name__ == '__main__':

  def __test1():
    # From branches/UQ/math/samples.py
    num_sample_points = 5
    lower = [10.0, 0.0, 2.1]
    upper = [100.0, 30.0, 2.8]

    pts = _random_samples(lower,upper,num_sample_points)
    print("randomly sampled points\nbetween %s and %s" % (lower, upper))
    print(pts)

  def __test2():
    # From branches/UQ/math/cut.py
    from mystic.tools import random_seed
    random_seed(123)
    lower = [-60.0, -10.0, -50.0]
    upper = [105.0, 30.0, 75.0]

    def model(x):
      x1,x2,x3 = x
      if x1 > (x2 + x3): return x1*x2 - x3
      return 0.0

    failure,success = sample(model,lower,upper)
    pof = float(failure) / float(failure + success)
    print("PoF using method 1: %s" % pof)
    random_seed(123)
    print("PoF using method 2: %s" % sampled_pof(model,lower,upper))

  # run the tests
  __test1()
  __test2()


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2022-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
#
# adapted from XRD example at: https://spotlight.readthedocs.io/
"""
optimization of 3-input cost function using online learning of a surrogate
"""
import os
from mystic.solvers import diffev2
from mystic.monitors import LoggingMonitor
from mystic.math.legacydata import datapoint
from mystic.cache.archive import file_archive, read as get_db
from ouq_models import WrapModel, LearnedModel
from emulators import cost3 as cost, x3 as target, bounds3 as bounds

# prepare truth (i.e. an 'expensive' model)
nx = 3; ny = None
archive = get_db('truth.db', type=file_archive)
truth = WrapModel("truth", cost, nx=nx, ny=ny, rnd=False, cached=archive)

# remove any prior cached evaluations of truth
archive.clear(); archive.sync(clear=True)
if os.path.exists("error.txt"): os.remove("error.txt")
if os.path.exists("log.txt"): os.remove("log.txt")

try: # parallel maps
    from pathos.maps import Map
    from pathos.pools import ProcessPool, ThreadPool, SerialPool
    pmap = Map(SerialPool) #ProcessPool
except ImportError:
    pmap = None

# generate a dataset by sampling truth
data = truth.sample(bounds, pts=[2, 1, 1], map=pmap)

# shutdown mapper
if pmap is not None:
    pmap.close(); pmap.join(); pmap.clear()

# create an inexpensive surrogate for truth
import torch
from torch.optim.lr_scheduler import StepLR
from skorch.callbacks.lr_scheduler import LRScheduler
from sklearn.preprocessing import StandardScaler
from constrained_skorch import LinearRegression, NetRegressor
lr_policy = LRScheduler(StepLR, step_size=15, gamma=0.5)
estimator = NetRegressor(LinearRegression, criterion=torch.nn.MSELoss,
                         train_split=None,
                         module__n_nx=nx, module__n_h=300,
                         module__n_layers=4, module__dropout=0.2,
                         optimizer=torch.optim.Adam, optimizer__lr=.005,
                         max_epochs=200, callbacks=[lr_policy],
                         device='cpu', batch_size=64, verbose=0)

from ml import Estimator, MLData, improve_score
kwds = dict(estimator=estimator, transform=StandardScaler())
# iteratively improve estimator
mlp = Estimator(**kwds)
best = improve_score(mlp, MLData(data.coords, data.coords, data.values, data.values), tries=10, verbose=True)
mlkw = dict(estimator=best.estimator, transform=best.transform, rnd=False)
surrogate = LearnedModel('surrogate', nx=nx, ny=ny, data=truth, **mlkw)

# iterate until error (of candidate minimum) < 1e-3
tracker = LoggingMonitor(1, filename='error.txt', label='error')
from mystic.abstract_solver import AbstractSolver
from mystic.termination import VTR
loop = AbstractSolver(nx)
loop.SetTermination(VTR(1e-2)) #XXX: VTRCOG, TimeLimits, etc?
loop.SetEvaluationLimits(maxiter=100)
loop.SetGenerationMonitor(tracker)
while not loop.Terminated():

    # fit the surrogate to data in truth database
    surrogate.fit(data=data)

    # find the minimum of the surrogate
    results = diffev2(lambda x: surrogate(x), bounds, npop=20,
                      bounds=bounds, gtol=500, full_output=True)

    # evaluate truth at the same input as the surrogate minimum
    xnew = results[0].tolist()
    ynew = truth(xnew)
    # add most recent candidate minimum evaluated with truth to database
    data.append(datapoint(xnew, value=ynew))

    # compute absolute error between truth and surrogate at candidate minimum
    ysur = results[1]
    error = abs(ynew - ysur)
    print("truth: %s @ %s" % (ynew, xnew))
    print("candidate: %s; error: %s" % (ysur, error))
    print("evaluations of truth: %s" % len(data))

    # save to tracker if less than current best
    ysave = error # track error when learning surrogate
    if len(tracker) and tracker.y[-1] < ysave:
        tracker(*tracker[-1])
    else: tracker(xnew, ysave)

# get the results at the best parameters from the truth database
xbest = tracker[-1][0]
ybest = archive[tuple(xbest)]

# print the best parameters
print(f"Best solution is {xbest} with beta {ybest}")
print(f"Reference solution: {target}")
ratios = [x / y for x, y in zip(target, xbest)]
print(f"Ratios of best to reference solution: {ratios}")



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
One mogi source, similar to test_mogi, but uses
CostFactory objects

"""

from test_mogi import *

from mystic.forward_model import CostFactory
from mystic.filters import component

def de_solve(CF):
    solver = DifferentialEvolutionSolver(ND, NP)

    stepmon = Monitor()
    minrange = [-1000., -1000., -100., -10.];
    maxrange = [1000., 1000., 100., 10.];
    solver.SetRandomInitialPoints(min = minrange, max = maxrange)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(stepmon)

    solver.Solve(CF, termination = ChangeOverGeneration(generations=100), \
                 CrossProbability=0.5, ScalingFactor=0.5)

    solution = solver.Solution()
  
    return solution, stepmon

if __name__ == '__main__':
    F = CostFactory()
    F.addModel(ForwardMogiFactory, 4, 'mogi1', outputFilter=component(2))
    myCostFunction = F.getCostFunction(evalpts=stations, observations=data_z)
    print(F)
    rp =  F.getRandomParams()
    print("new Cost Function : %s " % myCostFunction(rp))
    print("orig Cost Function: %s " % cost_function(rp))

    f1 = ForwardMogiFactory(rp)
    f2 = ForwardMogiFactory(rp)

    print('start cf')
    for i in range(3000):
        xx = cost_function(rp)
    print('end cf')

    print('start cf2')
    for i in range(3000):
        xx = myCostFunction(rp)
    print('end cf2')

    #de_solve(cost_function)
    de_solve(myCostFunction)

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2019-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
This module contains the base class for mystic samplers, and describes
the mystic sampler interface.

Example usage:

    >>> # the fuction to be sampled and the bounds
    >>> from mystic.models import rosen
    >>> lb = [0.0, 0.0, 0.0]
    >>> ub = [2.0, 2.0, 2.0]
    >>> bounds = list(zip(lb, ub))
    >>>
    >>> # instantiate and configure the nested solver
    >>> from mystic.solvers import NelderMeadSimplexSolver
    >>> solver = NelderMeadSimplexSolver(3)
    >>>
    >>> # instantiate and configure the sampler
    >>> from mystic.samplers import SparsitySampler
    >>> sampler = SparsitySampler(bounds, rosen, npts=4, solver=solver)
    >>>
    >>> # sample `npts` without the solver
    >>> sampler.sample()
    >>> sampler.evals(all=True)
    [1, 1, 1, 1]
    >>>
    >>> # sample `npts` until all solvers terminate
    >>> sampler.terminated(all=True)
    [False, False, False, False]
    >>> sampler.sample_until(terminated=all)
    >>> sampler.evals(all=True)
    [323, 120, 132, 138]
    >>> sampler.terminated(all=True)
    [True, True, True, True]
    >>>
    >>> # extract the results
    >>> import numpy as np
    >>> np.array(sampler.bestEnergy(all=True))
    array([2.64786360e-10, 2.00199994e+03, 4.02369987e+02, 1.10040095e+00])
    >>> np.array(sampler.bestSolution(all=True))
    array([[1.00000555e+00, 1.00001074e+00, 1.00002251e+00],
           [1.99999999e+00, 2.09709611e-08, 1.99999999e+00],
           [1.45188568e+00, 1.99999999e+00, 2.00000000e+00],
           [4.45889175e-01, 1.92620174e-01, 1.12598488e-08]])
    >>> 

"""
__all__ = ['AbstractSampler']


class AbstractSampler(object): # derived class per sampling algorithm
    """
AbstractSampler base class for optimizer-directed samplers
    """
    def __init__(self, bounds, model, npts=None, **kwds):
        """
Args:
  bounds (bounds, default=None): the bounds for each parameter, given as
    a mystic.bounds instance or a list of ``(lower, upper)`` tuples.
  model (function): ``y = model(x)``, where ``x`` is an input parameter vector
  npts (int, default=None): number of points to sample the model
  **kwds (dict, default={}): keywords for the underlying ensemble of solvers;
    ``(evalmon, stepmon, maxiter, maxfun, dist, saveiter, state, termination,
    constraints, penalty, reducer, solver, id, map, tightrange, cliprange)``
    are available for use. See ``mystic.ensemble`` for more details.
        """ #FIXME: added 'id' and 'map'
        self._bounds = bounds
        self._model = model #FIXME: if None, interpolate
        self._npts = npts

        self._evalmon = kwds.pop('evalmon', None)
        if self._evalmon is None:
            from mystic.monitors import Null
            self._evalmon = Null()
        from collections import defaultdict
        self._kwds = defaultdict(lambda :None)
        self._kwds.update(kwds)

        kwd = dict(tight=self._kwds['tightrange'], clip=self._kwds['cliprange'])
        s = self._init_solver(**kwd)
        #s.SetStrictRanges(*zip(*bounds), **kwd)
        #s.SetObjective(model) #FIXME: ensure cached model
        ##s.SetObjective(memo) #model) #XXX: ExtraArgs: axis ???

        # apply additional kwds
        solver = self._kwds['solver']
        if solver is not None:
            if 'NP' in self._kwds: #XXX: instead use npts?
                s.SetNestedSolver(solver, NP=self._kwds['NP'])
            s.SetNestedSolver(solver)
        s.id = self._kwds['id']
        s.SetDistribution(self._kwds['dist'])
        s.SetEvaluationLimits(self._kwds['maxiter'], self._kwds['maxfun'])
        s.SetSaveFrequency(self._kwds['saveiter'], self._kwds['state'])
        termination = self._kwds['termination']
        if termination is not None: s.SetTermination(termination) #XXX:?
        s.SetConstraints(self._kwds['constraints'])
        s.SetPenalty(self._kwds['penalty'])
        s.SetReducer(self._kwds['reducer'], arraylike=True)
        s.SetGenerationMonitor(self._kwds['stepmon']) #XXX: use self._stepmon?
        mapper = self._kwds['map']
        if mapper is not None: s.SetMapper(mapper)
        #TODO: close/join/clear after Solve?

        # pass a copy of the monitor to all instances
        import copy
        m = copy.deepcopy(self._evalmon) #XXX: no change with direct reference
        s.SetEvaluationMonitor(m)
        self._sampler = s
        self._npts = s._npts

        self._evals = [0] * s._npts #XXX: collect count or monitor?
        self._iters = [0] * s._npts #XXX: collect count or monitor?
        return


    def _init_solver(self, **kwd):
        """initialize the ensemble solver"""
        from mystic.abstract_ensemble_solver import AbstractEnsembleSolver
        s = AbstractEnsembleSolver(len(self._bounds), npts=self._npts)
        s.SetStrictRanges(*zip(*self._bounds), **kwd)
        s.SetObjective(self._model)
        return s


    def _reset_sampler(self):
        """reinitialize the ensemble solver/sampler"""
        s = self._sampler
        # the following "seems" to work; all/more necessary?
        #XXX: add this method as internal method for ensemble solver?
        s.popEnergy = [s._init_popEnergy] * s.nPop
        s.population = [[0.0 for i in range(s.nDim)] for j in range(s.nPop)]
        # s.trialSolution = [0.0] * s.nDim
        s._bestEnergy = None
        s._bestSolution = None
        s._fcalls = [0] #XXX:?
        s._allSolvers = [None] * s._npts
        s._bestSolver = None
        s._init_solution = [None] * s._npts
        s.SetGenerationMonitor(s._stepmon[:0], new=True) #XXX:?
        s.SetEvaluationMonitor(s._evalmon[:0], new=True) #XXX:?
        s._live = False
        s._bootstrap_objective()
        return


    def _reset_solved(self):
        """reinitialize all terminated solvers in the ensemble"""
        s = self._sampler
        # the following "seems" to work; all/more necessary?
        #XXX: add this method as internal method for ensemble solver?
        from copy import copy as _copy
        solver = s._AbstractEnsembleSolver__get_solver_instance(reset=True)
        #FIXME: potentially do _InitialPoints, but only where Terminated
        #TODO: test impact of resetting monitor (behavior was reset=False)
        [s._allSolvers.__setitem__(i, _copy(solver)) for i,j in enumerate(s.Terminated(all=True)) if j is True]
        s._bestSolver = None
        s._AbstractEnsembleSolver__update_state()
        return


    def _sample(self, reset=False):#, **kwds):
        """collect a sample for each member in the ensemble

Args:
  reset (bool, default=False): reset all solvers before sampling; alternately,
    if ``reset`` is None, then only reset the terminated solvers

Returns:
  None
        """
        #NOTE: sample(reset_all) is like _sample(reset), with None <=> False
        s = self._sampler
        m = self._evalmon

        _eval = s._all_evals
        _iter = s._all_iters
        if reset: self._reset_sampler()
        elif reset is None: self._reset_solved()

        # get evals before taking a step (needed for backfill)
        evals = [len(getattr(i, '_evalmon', ())) for i in s._allSolvers]
        neval = len(m)

        # take a step
        s.Step()
        termination = None #s.Terminated(**kwds) #XXX:?
        #NOTE: i._evalmon._id for i in _allSolvers is [None]

        # backfill original monitor instance #NOTE: not in chronological order
        map(m.extend, [i._evalmon[e:] for e,i in zip(evals, s._allSolvers)])

        _eval = (max(j-i,1) if j>=i else i for (i,j) in zip(_eval,s._all_evals))
        _iter = (max(j-i,1) if j>=i else i for (i,j) in zip(_iter,s._all_iters))
        for i,(e,t) in enumerate(zip(_eval,_iter)):
            self._evals[i] += e
            self._iters[i] += t
        return termination


    def sample(self, if_terminated=None, reset_all=True):#, **kwds):
        """sample npts using vectorized solvers

Args:
  if_terminated (bool, default=None): the amount of termination; must be one
    of ``{all, True, any, False, None}``
  reset_all (bool, default=True): action to take when ``if_terminated`` is met;
    must be one of ``{True, False, None}``

Returns:
  None

Notes:
  - When ``if_terminated`` is all, reset if all solvers are terminated.
  - When ``if_terminated`` is True, reset if the best solver is terminated.
  - When ``if_terminated`` is any, reset if any solvers are terminated.
  - When ``if_terminated`` is False, reset if no solvers are terminated.
  - When ``if_terminated`` is None, always reset, unless ``reset_all`` is None.
  - If ``reset_all`` is True, reset all solvers if ``if_terminated`` is met.
  - If ``reset_all`` is False, similarly reset only the terminated solvers.
  - If ``reset_all`` is None, never reset (thus ignores ``if_terminated``).
        """
        if repr(reset_all) in ('True','False'):
            s = self._sampler
            if if_terminated is None: # reset regardless
                return self._sample(reset=(True if reset_all else None))#, **kwds)
            if not if_terminated in (all, True, any, False):
                msg = "%s not in (all, True, any, False, None)" % if_terminated
                raise ValueError(msg)
            stop = s.Terminated(all=True)
            if all(stop) \
            or (stop[s._is_best() or 0] and if_terminated is not all) \
            or (any(stop) and if_terminated in (any, False)) \
            or (not any(stop) and if_terminated is False):
                return self._sample(reset=(True if reset_all else None))#, **kwds)
            # othewise just take a step
            return self._sample(reset=False)#, **kwds)
        return self._sample(reset=False)#, **kwds)


    def _sample_until(self, iters=None, evals=None, terminated=None, **kwds):
        """sample until one of the stop conditions are met"""
        from numpy import inf, all as all_, any as any_
        if evals is None:
            evals = inf
        if iters is None:
            iters = inf
        if terminated is None:
            terminated = self._npts + 1
        else: #NOTE: if termination is specified, don't reset by default
            kwds['reset_all'] = kwds.get('reset_all', None)
            if terminated is all or terminated is all_:
                terminated = self._npts
            elif terminated is any or terminated is any_:
                terminated = 1
            elif terminated is True: # best
                pass
            elif terminated is False:
                terminated = 0
            elif terminated > self._npts:
                msg = 'terminated = %s is greater than npts = %s' % (terminated, self._npts)
                raise ValueError(msg)
        if_stop = kwds.pop('if_terminated', None)
        quit = inf
        if if_stop is all or if_stop is all_:
            quit = self._npts
        elif if_stop is any or if_stop is any_:
            quit = 1
        elif if_stop is True: # best
            quit = int(True)
        elif if_stop is False:
            quit = 0
        elif if_stop is None:
            quit = 0 #XXX: -1 ?
        reset = kwds.pop('reset_all', True)
        # require iters or evals if: (term > npts) OR (reset!=None & quit<term)
        if iters is inf and evals is inf:
            msg = ', evals or iters must be < inf'
            if terminated > self._npts:
                msg = ('if terminated > %s' % self._npts) + msg
                raise ValueError(msg)
            elif (reset is not None and quit < terminated):
                msg = ('if terminated > %s, reset_all = %s' % (quit,reset)) + msg
                raise ValueError(msg)
        return (iters, evals, terminated, if_stop, reset)

    def sample_until(self, iters=None, evals=None, terminated=None, **kwds):
        """sample until one of the stop conditions are met

Args:
  iters (int, default=inf): maximum number of iterations
  evals (int, default=inf): maximum number of evaluations
  terminated (int, default=inf): maximum number of terminated solvers
  if_terminated (bool, default=None): the amount of termination; must be one
    of ``{all, True, any, False, None}``
  reset_all (bool, default=True): action to take when ``if_terminated`` is met;
    must be one of ``{True, False, None}``

Notes:
  Possible stop conditions are:
  - solver iterations ``iters()`` equals or exceeds ``iters``
  - solver evaluations ``evals()`` equals or exceeds ``evals``
  - number of terminated solvers equals or exceeds ``terminated``

Notes:
  When sampling stops, the reset conditions are:
  - ``if_terminated` equals or exceeds the number of terminated solvers
  - ``reset_all` indicates which solvers to reset

Notes:
  - At least one of ``(iters, evals, termination)`` must be less than 'inf'.
  - ``terminated`` may also be one of ``{all, True, any, False, None}``, where
    ``{all: 'all', True: 'best', any: '1', False: '0', None: 'inf', N: 'N'}``
  - ``if_terminated`` may be one of ``{all, True, any, False, None}``, where
    ``{all: 'all', True: 'best', any: '1', False: '0', None: 'always'}``
  - ``reset_all`` may be one of ``{True, False, None}``, where ``{True: 'reset
    all', False: 'reset solved', None: 'never reset'}``

Notes: ### ??? ??? ###
  - The default sampler configuration is to always reset (``reset_all=True``)
  - If ``termination != None``, the default is never reset (``reset_all=None``)
        """
        args = self._sample_until(iters, evals, terminated, **kwds)
        (iters, evals, terminated, if_stop, reset) = args
        ###stop = self.terminated(**kwds)
        while self.iters() < iters and self.evals() < evals and (not self.terminated(all=False) if (terminated is True) else (sum(self.terminated(all=True)) < terminated)): #FIXME: sometimes limiting iters stops very late
            stop = self.sample(if_stop, reset)#, **kwds)
        return### stop


    def evals(self, all=False): #XXX: None?
        """get the total number of function evaluations

Args:
  all (bool, default=False): report the ``evals`` for each ensemble member
        """
        if all:
            return self._evals
        return sum(self._evals)


    def iters(self, all=False): #XXX: None?
        """get the total number of solver iterations

Args:
  all (bool, default=False): report the ``iters`` for each ensemble member
        """
        if all:
            return self._iters
        return sum(self._iters)


    def bestEnergy(self, all=False): #XXX: None?
        """get the best sampled Energy

Args:
  all (bool, default=False): report the ``energy`` for each ensemble member
        """
        if all:
            return self._sampler._all_bestEnergy
        return self._sampler.bestEnergy


    def bestSolution(self, all=False): #XXX: None?
        """get the best sampled solution

Args:
  all (bool, default=False): report the ``solution`` for each ensemble member
        """
        from mystic.tools import listify as tolist
        if all:
            return tolist(self._sampler._all_bestSolution)
        return tolist(self._sampler.bestSolution)


    def initEnergy(self, map=None): #XXX: None?
        """sample the model at the initial solution (without running the solver)

Args:
  all (map, default=None): map function instance
        """
        if map is None:
            from builtins import map
        return list(map(self._model, self._sampler._init_solution))


    def initSolution(self):
        """get the initial solution (without running the solver)
        """
        from mystic.tools import listify as tolist
        return tolist(self._sampler._init_solution)


    def terminated(self, *args, **kwds): #FIXME: confusing wrt if_terminated?
        """check if termination conditions have been met

Args:
  disp (bool, default=False): print termination statistics and/or warnings
  info (bool, default=False): return termination message (instead of boolean)
  all (bool, default=None): get results for all solvers, else get the 'best'

Notes:
  - ``disp`` expects a bool, but can also take ``'verbose'`` for more verbosity 
  - ``all``, by default (i.e. None), will show only the terminated solvers
        """
        _all = kwds.get('all', None)
        if _all not in (True, False, None):
            msg = 'all = %s, must be one of (True, False, None)' % getattr(_all, '__name__', _all)
            raise ValueError(msg)
        kwds['all'] = _all
        return self._sampler.Terminated(*args, **kwds)


"""
### Notes ###

# utility functions
_reset_sampler: reset all
_reset_solved: reset all solved
_sample(reset): reset all/stop/none, then step

# use cases to manage evals/iters
_restart_all_without_step: _reset_sampler
_restart_solved_without_step: _reset_solved
_restart_all_then_step: _sample(reset=True)
_restart_solved_then_step: _sample(reset=None)
_step_without_restart: _sample(reset=False)

# options
_restart_if_solved(solved): (* = 'all' if all else 'solved')
. all: if all stop, _restart_*_then_step
. True: if best stop, _restart_*_then_step
. any: if any stop, _restart_*_then_step 
. False: if none stop, _restart_*_then_step
. None: _step_without_restart
"""



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2019-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''klepto archive readers and writers, for functions and data
'''
from klepto.archives import (dict_archive, dir_archive, file_archive,
                             hdf_archive, hdfdir_archive, null_archive,
                             sql_archive, sqltable_archive)

def read(name, keys=None, keymap=None, type=None):
    """read klepto db with name 'name'

    Args:
        name (string): filename of the klepto db
        keys (iterable): keys to load; or boolean to load all/no keys
        keymap (klepto.keymap): used for key encoding
        type (klepto.archive): type of klepto archive

    Returns:
        klepto db object (has dictionary-like interface)

    Notes:
        If keys is None, create a direct handle to the db.
        If a key in keys is not found, it will be ignored.
    """
    if type is None:
        from klepto.archives import dir_archive as _archive
    else:
        _archive = type
    if keys is None:
        return _archive(name, cached=False)
    if keymap is None:
        from klepto.keymaps import keymap as _kmap
        keymap = _kmap()
    archive = _archive(name, cached=True)
    if keys is False:
        return archive
    if keys is True:
        archive.load()
    else: # apply keymap to keys
        archive.load(*(keymap(*(k if hasattr(k, '__len__') else (k,))) for k in keys))
    return archive # archive with keymapped keys


def write(archive, entries, keymap=None):
    """write entries to klepto db instance 'archive'

    Args:
        archive (klepto.archive): archive instance
        entries (iterable): iterable/dict to update the archive with
        keymap (klepto.keymap): keymap used for key encoding

    Returns:
        None

    Examples:
        >>> write(foo, {x0:y0, x1:y1, ...})
        >>> write(foo, ((x0,y0), (x1,y1), ...))

    Notes:
        Within ``entries``, ``xi`` is a tuple of floats, and ``yi`` is a float.
    """
    if isinstance(archive, (str, (u'').__class__)):
        archive = read(archive, keymap=keymap) # assumes default type
    if keymap is not None: # apply keymap to keys
        entries = {keymap(*(k if hasattr(k, '__len__') else (k,))):v for k, v in entries.items()}
    # archive = read(name, keys=(False if cached else None))
    cached = archive.archive is not archive
    archive.update(entries) #XXX: for non-cached, __setitem__ is faster?
    if cached: archive.dump()
    return


def _read_func(name, keymap=None, type=None):
    """read function db with name 'name'

    Args:
        name (string): filename of the klepto db
        keymap (klepto.keymap): keymap used for key encoding
        type (klepto.archive): type of klepto archive

    Returns:
        klepto db object (has dictionary-like interface)
    """
    return read(name, keymap=keymap, type=type)


def read_func(name, keymap=None, type=None, n=0):
    """read stored function from db with name 'name'

    Args:
        name (string): filename of the klepto db
        keymap (klepto.keymap): keymap used for key encoding
        type (klepto.archive): type of klepto archive
        n (int): db entry in reverse order (i.e. most recent is ``0``)

    Returns:
        tuple of (stored function, distance information)

    Notes:
        If the db is empty, or ``n`` produces a bad index, returns ``None``.
        Alternately, ``name`` can be the relevant klepto.archive instance.
    """
    if not isinstance(name, (str, (u'').__class__)):
        if type is not None:
           #msg = 'if a klepto.archive instance is provided, type must be None'
           #raise TypeError(msg)
           type = None #NOTE: ignore type
        archive = getattr(name, '__archive__', name) # need cached == False
    else:
        archive = _read_func(name, type=type) # read entire archive to get size
    size = len(archive) - n - 1
    entry = size if keymap is None else keymap(size)
    return archive.get(entry, None) # apply keymap


def _write_func(archive, func, dist, keymap=None):
    """write stored function and distance information to archive

    Args:
        archive (klepto.archive): function archive (output of ``_read_func``)
        func (function): with interface ``y = f(x)``, ``x`` is a list of floats
        dist (dict): distance information
        keymap (klepto.keymap): keymap used for key encoding

    Returns:
        None
    """
    arxiv = getattr(archive, '__archive__', archive)
    entries = {len(arxiv): (func, dist)}
    write(archive, entries, keymap=keymap) # apply keymap to keys
    return


def get_dist(archive, func, keymap=None):
    """get the graphical distance of func from data in archive

    Args:
        archive (klepto.archive): run archive (output of ``read``)
        func (function): with interface ``y = f(x)``, ``x`` is a list of floats
        keymap (klepto.keymap): keymap used for key encoding

    Returns:
        array of floats, graphical distance from func to each point in archive
    """
    from mystic.math.legacydata import dataset
    data = dataset()
    arxiv = getattr(archive, '__archive__', archive)
    if not len(arxiv):
        return
    #FIXME: need to implement generalized inverse of keymap
    #HACK: only works in certain special cases
    kind = getattr(keymap, '__stub__', '')
    if kind in ('encoding', 'algorithm'): #FIXME always assumes used repr
        inv = lambda k: eval(k)
    elif kind in ('serializer', ): #FIXME: ignores all config options
        import importlib
        inv = lambda k: importlib.import_module(keymap.__type__).loads(k)
    else: #FIXME: give up, ignore keymap
        inv = lambda k: k
    y = ((inv(k), v) for k, v in arxiv.items())
    #HACK end
    import numpy as np
    y = np.array(list(y), dtype=object).T
    data.load(y[0].tolist(), y[1].tolist(), ids=list(range(len(y[0]))))
    from mystic.math.distance import graphical_distance
    return graphical_distance(func, data)


def _prep_dist(distances):
    """create a dict of mean and max graphical distance

    Args:
        distances (array[float]): indicates the graphical distance

    Returns:
        ``dict(mean=distances.mean(), max=distances.max())``
    """
    if distances is None:
        import numpy as np
        return dict(mean=np.nan, max=np.nan)
    return dict(mean=distances.mean(), max=distances.max())


'''
### tests ###
def _test(fundb, sql=False):
    """read function database, returning tuple of (function, distance)

    Args:
        fundb (string): function db name
        sql (bool): True if fundb is a ``sql_archive`` (w/ ``stringmap``)

    Returns:
        tuple of (stored function, distance information)
    """
    if sql:
        from klepto.keymaps import stringmap as _keymap
        from klepto.archives import sql_archive as ar
        km = _keymap()
    else:
        km = ar = None
        from klepto.archives import hdf_archive as ar
    return read_func(fundb, keymap=km, type=ar)



def _make_test(func, rundb, fundb, runsql=False, funsql=False):
    """create/increment function sql database

    Args:
        func (function): with interface ``y = f(x)``, ``x`` is a list of floats
        rundb (string): name of the klepto db archive of runs
        fundb (string): name of the klepto function archive
        runsql (bool): True if rundb is a ``sql_archive`` (w/ ``stringmap``)
        funsql (bool): True if fundb is a ``sql_archive`` (w/ ``stringmap``)

    Returns:
        None
    """
    if runsql:
        from klepto.keymaps import stringmap as _keymap
        from klepto.archives import sql_archive as r_ar
        r_km = _keymap()
    else:
        r_km = r_ar = None
    if funsql:
        from klepto.keymaps import stringmap as _keymap
        from klepto.archives import sql_archive as f_ar
        f_km = _keymap()
    else:
        f_km = f_ar = None
        from klepto.archives import hdf_archive as f_ar
    db = read(rundb, keymap=r_km, type=r_ar)
    dist = _prep_dist(get_dist(db, func, keymap=r_km))
    db = _read_func(fundb, keymap=f_km, type=f_ar)
    _write_func(db, func, dist, keymap=f_km)
    return


def _dump(obj, path):
    """serialize obj to file located at path

    Args:
        obj (object): any python object
        path (string): path of new pickle file
    """
    import dill
    dill.dump(obj, open(path, 'wb'))


def _load(path):
    """load serializade obj from file located at path

    Args:
        path (string): path of existing pickle file
    """
    import dill
    return dill.load(open(path, 'rb'))
'''



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Solve 8th-order Chebyshev polynomial coefficients with DE.
    - Plot (x2) of convergence to Chebyshev polynomial.
    - Monitor (x2) Chi-Squared for Chebyshev polynomial.

Demonstrates:
    - standard models
    - expanded solver interface
    - built-in random initial guess
    - customized monitors and termination conditions
    - customized DE mutation strategies
    - use of solver members to retrieve results information
"""

# Differential Evolution solver
from mystic.solvers import DifferentialEvolutionSolver2

# Chebyshev polynomial and cost function
from mystic.models.poly import chebyshev8, chebyshev8cost
from mystic.models.poly import chebyshev8coeffs

# tools
from mystic.termination import VTR
from mystic.strategy import Best1Exp
from mystic.monitors import VerboseMonitor, Monitor
from mystic.tools import getch, random_seed
from mystic.math import poly1d
import matplotlib.pyplot as plt
plt.ion()

# draw the plot
def plot_frame(label=None):
    plt.close()
    plt.title("8th-order Chebyshev coefficient convergence")
    plt.xlabel("Differential Evolution %s" % label)
    plt.ylabel("Chi-Squared")
    plt.draw()
    plt.pause(0.001)
    return
 
# plot the polynomial trajectories
def plot_params(monitor):
    x = list(range(len(monitor)))
    y = monitor.y
    plt.plot(x,y,'b-')
    plt.axis([1,0.5*x[-1],0,y[1]])#,'k-')
    plt.draw()
    plt.pause(0.001)
    return


if __name__ == '__main__':

    print("Differential Evolution")
    print("======================")

    # set range for random initial guess
    ndim = 9
    x0 = [(-100,100)]*ndim
    random_seed(123)

    # configure monitors
    stepmon = VerboseMonitor(50)
    evalmon = Monitor()

    # use DE to solve 8th-order Chebyshev coefficients
    npop = 10*ndim
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=[-100]*ndim, max=[100]*ndim)
    solver.SetEvaluationLimits(generations=999)
    solver.SetEvaluationMonitor(evalmon)
    solver.SetGenerationMonitor(stepmon)
    solver.enable_signal_handler()
    solver.Solve(chebyshev8cost, termination=VTR(0.01), strategy=Best1Exp, \
                 CrossProbability=1.0, ScalingFactor=0.9)
    solution = solver.bestSolution

    # get solved coefficients and Chi-Squared (from solver members)
    iterations = solver.generations
    cost = solver.bestEnergy
    print("Generation %d has best Chi-Squared: %s" % (iterations, cost))
    print("Solved Coefficients:\n %s\n" % poly1d(solver.bestSolution))

    # plot convergence of coefficients per iteration
    plot_frame('iterations')
    plot_params(stepmon)
    getch()

    # plot convergence of coefficients per function call
    plot_frame('function calls')
    plot_params(evalmon)
    getch()

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Mogi's model of surface displacements from a point spherical source in an
elastic half space

References:
    1. Mogi, K. "Relations between the eruptions of various
       volcanoes and the deformations of the ground surfaces around them", 
       Bull. Earthquake. Res. Inst., 36, 99-134, 1958.
"""
__all__ = ['Mogi', 'mogi']

from .abstract_model import AbstractModel

from numpy import sum as numpysum
from numpy import array, pi

class Mogi(AbstractModel):
    """
Computes surface displacements Ux, Uy, Uz in meters from a point spherical
pressure source in an elastic half space [1].
    """

    def __init__(self,name='mogi',metric=lambda x: numpysum(x*x),sigma=1.0):
        AbstractModel.__init__(self,name,metric,sigma)
        return

    def evaluate(self,coeffs,evalpts):
        """evaluate a single Mogi peak over a 2D (2 by N) numpy array of evalpts, where coeffs = (x0,y0,z0,dV)"""
        x0,y0,z0,dV = coeffs
        dx = evalpts[0,:] - x0
        dy = evalpts[1,:] - y0
        dz = 0 - z0
        c = dV * 3. / 4. * pi
        # or equivalently c= (3/4) a^3 dP / rigidity
        # where a = sphere radius, dP = delta Pressure
        r2 = dx*dx + dy*dy + dz*dz
        C = c / pow(r2, 1.5)
        return array((C*dx,C*dy,C*dz)) #XXX: requires a numpy.array

    def ForwardFactory(self,coeffs):
        """generates a mogi source instance from a list of coefficients"""
        x0,y0,z0,dV = coeffs
        def forward_mogi(evalpts):
            """a single Mogi peak over a 2D (2 by N) numpy array
with (x0,y0,z0,dV) = (%s,%s,%s,%s)""" % (x0,y0,z0,dV)
            return self.evaluate((x0,y0,z0,dV),evalpts)
        return forward_mogi

    #FIXME: continue refactoring from test_mogi*.py...
    def CostFactory(self,target,pts):
        """generates a cost function instance from list of coefficients & evaluation points"""
        raise NotImplementedError("cost function not implemented")

    #FIXME: continue refactoring from test_mogi*.py...
    def CostFactory2(self,pts,datapts,nparams):
        """generates a cost function instance from datapoints & evaluation points"""
        raise NotImplementedError("cost function not implemented")

    pass
 

# prepared instances
mogi = Mogi()

# End of file



#!/usr/bin/env python
#
# Problem definition:
# Example in google/or-tools
# https://github.com/google/or-tools/blob/master/examples/python/integer_programming.py
# with Copyright 2010-2013 Google
# and disclamer as stated at the above reference link.
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from integer_programming import objective, bounds, xs, ys
# bounds = [(0,11)]*2  #XXX: MOD = range(11) instead of LARGE

from mystic.penalty import quadratic_inequality, quadratic_equality
from mystic.constraints import as_constraint, discrete

def penalty1(x): # <= 0.0
    return -3*x[0] - 2*x[1] + 17.0

def penalty2(x): # == 0.0
    from numpy import abs, round
    return abs(x - round(x)).sum() # penalize when not an 'int'

#@quadratic_equality(penalty2)
@quadratic_inequality(penalty1)
def penalty(x):
    return 0.0

solver = as_constraint(penalty)
#solver = discrete(range(11))(solver)  #XXX: MOD = range(11) instead of LARGE
#FIXME: constrain to 'int' with discrete is very fragile!  required #MODs

def constraint(x):
    from numpy import round
    return round(solver(x))

# better is to constrain to integers, penalize otherwise
from mystic.constraints import integers

@integers()
def round(x):
  return x


if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, penalty=penalty, constraints=round, npop=30, gtol=50, disp=True, full_output=True)

    print(result[0])
    assert almostEqual(result[0], xs, tol=1e-8) #XXX: fails b/c rel & zero?
    assert almostEqual(result[1], ys, tol=1e-4)


# EOF



#!/usr/bin/env python
#
# Problem definition and original response:
# https://stackoverflow.com/q/21765794/2379433
# https://stackoverflow.com/a/43162017/2379433
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2018-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
solve:
     max [5th percentile of (ui_T*X), i in 1 to M]

such that: 
     0<=X<=1,
     [95th percentile of (X_T*si*X), i in 1 to M]<= constant

with:
     unknowns X (Nx1), M (Nx1) u vectors, and M (NxN) s matrices.
'''
import numpy as np

M = 10
N = 3
Q = 10
C = 10

# let's be lazy, and generate s and u randomly...
s = np.random.randint(-Q,Q, size=(M,N,N))
u = np.random.randint(-Q,Q, size=(M,N))

def percentile(p, x):
    x = np.sort(x)
    p = 0.01 * p * len(x)
    if int(p) != p:
        return x[int(np.floor(p))]
    p = int(p)
    return x[p:p+2].mean()

def objective(x, p=5): # inverted objective, to find the max
    return -1*percentile(p, [np.dot(np.atleast_2d(u[i]), x)[0] for i in range(0,M-1)])


def constraint(x, p=95, v=C): # 95%(xTsx) - v <= 0
    x = np.atleast_2d(x)
    return percentile(p, [np.dot(np.dot(x,s[i]),x.T)[0,0] for i in range(0,M-1)]) - v

from mystic.bounds import Bounds
bounds = Bounds(0,1,n=N)

from mystic.penalty import quadratic_inequality
@quadratic_inequality(constraint, k=1e4)
def penalty(x):
  return 0.0

from mystic.solvers import diffev2
from mystic.monitors import VerboseMonitor
mon = VerboseMonitor(10)

result = diffev2(objective, x0=bounds, penalty=penalty, npop=10, gtol=200, \
                 disp=False, full_output=True, itermon=mon, maxiter=M*N*100)

print(result[0])
print(result[1])



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

"""
Tests of factories that enable verbose and compound termination conditions
"""
disp = 0

def __and(*funcs):
  "factory for compounding termination conditions with *and*"
  def term(solver, info=False): #FIXME: 'self' doesn't flatten
    stop = {}
    [stop.update({f : f(solver, info)}) for f in funcs]
    _all = all(stop.values())
    if not info: return _all
    if info == 'self': return tuple(stop.keys()) if _all else ()
    return "; ".join(set("; ".join(stop.values()).split("; "))) if _all else ""
  return term

def __or(*funcs):
  "factory for compounding termination conditions with *or*"
  def term(solver, info=False): #FIXME: 'self' doesn't flatten
    stop = {}
    [stop.update({f : f(solver, info)}) for f in funcs]
    _any = any(stop.values())
    if not info: return _any
    [stop.pop(cond) for (cond,met) in stop.items() if not met]
    if info == 'self': return tuple(stop.keys())
    return "; ".join(set("; ".join(stop.values()).split("; ")))
  return term

def __when(func):
  "factory for single termination condition cast as a compound condition"
  return __and(func)


from mystic.termination import And, Or, When


def test_me(info='self'):
    from mystic.solvers import DifferentialEvolutionSolver
    s = DifferentialEvolutionSolver(4,4)

    from mystic.termination import VTR
    from mystic.termination import ChangeOverGeneration
    from mystic.termination import NormalizedChangeOverGeneration
    v = VTR()
    c = ChangeOverGeneration()
    n = NormalizedChangeOverGeneration()

    if disp: print("define conditions...")
    _v = When(v)
    _c = When(c)
    _n = When(n)
    v_and_c = And(v,c)
    v_and_n = And(v,n)
    v_or_c = Or(v,c)
    v_or_n = Or(v,n)
    v_and_c__or_n = Or(v_and_c,_n)
    v_and_c__or_c = Or(v_and_c,_c)
    v_and_n__or_n = Or(v_and_n,_n)
    c_and_v = And(c,v)
    c_or_v = Or(c,v)
    c_or__c_and_v = Or(_c,c_and_v)

    assert len(_v) == len(_c) == len(_n) == 1
    assert list(_v).index(v) == list(_c).index(c) == list(_n).index(n) == 0
    assert list(_v).count(v) == list(_c).count(c) == list(_n).count(n) == 1
    assert v in _v and c in _c and n in _n
    assert v in v_and_c and c in v_and_c
    assert v in v_and_n and n in v_and_n
    assert v in v_or_c and c in v_or_c
    assert v in v_or_n and n in v_or_n
    assert len(v_and_c) == len(v_and_n) == len(v_or_c) == len(v_or_n) == 2
    assert len(v_and_c__or_n) == len(v_and_c__or_c) == len(v_and_n__or_n) == 2
    assert v not in v_and_c__or_n and c not in v_and_c__or_n and n not in v_and_c__or_n
    assert v_and_c in v_and_c__or_n and _n in v_and_c__or_n
    assert v_and_c in v_and_c__or_c and _c in v_and_c__or_c
    assert v_and_n in v_and_n__or_n and _n in v_and_n__or_n
    assert c in c_and_v and v in c_and_v
    assert c in c_or_v and v in c_or_v
    assert list(c_and_v).index(c) == list(v_and_c).index(v)
    assert list(c_and_v).index(v) == list(v_and_c).index(c)
    assert list(c_or_v).index(c) == list(v_or_c).index(v)
    assert list(c_or_v).index(v) == list(v_or_c).index(c)
    assert c_and_v in c_or__c_and_v and _c in c_or__c_and_v

    if disp:
        print("_v:%s" % _v)
        print("_c:%s" % _c)
        print("_n:%s" % _n)
        print("v_and_c:%s" % v_and_c)
        print("v_and_n:%s" % v_and_n)
        print("v_or_c:%s" % v_or_c)
        print("v_or_n:%s" % v_or_n)
        print("v_and_c__or_n:%s" % v_and_c__or_n)
        print("v_and_c__or_c:%s" % v_and_c__or_c)
        print("v_and_n__or_n:%s" % v_and_n__or_n)
        print("c_and_v:%s" % c_and_v)
        print("c_or_v:%s" % c_or_v)
        print("c_or__c_and_v:%s" % c_or__c_and_v)

    if disp: print("initial conditions...")
    _v_and_c = v_and_c(s, info)
    _v_and_n = v_and_n(s, info)
    _v_or_c = v_or_c(s, info)
    _v_or_n = v_or_n(s, info)

    assert not _v_and_c
    assert not _v_and_n
    assert not _v_or_c
    assert not _v_or_n

    if disp:
        print("v_and_c:%s" % _v_and_c)
        print("v_and_n:%s" % _v_and_n)
        print("v_or_c:%s" % _v_or_c)
        print("v_or_n:%s" % _v_or_n)

    if disp: print("after convergence toward zero...")
    s.energy_history = [0,0,0,0,0,0,0,0,0,0,0,0]
    # _v and _n are True, _c is False
    _v_and_c = v_and_c(s, info)
    _v_and_n = v_and_n(s, info)
    _v_or_c = v_or_c(s, info)
    _v_or_n = v_or_n(s, info)

    assert not _v_and_c
    assert _v_and_n
    assert _v_or_c
    assert _v_or_n
    if info == 'self':
        assert v in _v_and_n and n in _v_and_n
        assert v in _v_or_c and c not in _v_or_c
        assert v in _v_or_n and n in _v_or_n

    if disp:
        print("v_and_c:%s" % _v_and_c)
        print("v_and_n:%s" % _v_and_n)
        print("v_or_c:%s" % _v_or_c)
        print("v_or_n:%s" % _v_or_n)

    if disp: print("nested compound termination...")
    # _v and _n are True, _c is False
    _v_and_c__or_n = v_and_c__or_n(s, info)
    _v_and_c__or_c = v_and_c__or_c(s, info)
    _v_and_n__or_n = v_and_n__or_n(s, info)

    assert _v_and_c__or_n
    assert not _v_and_c__or_c
    assert _v_and_n__or_n
    if info == 'self':
        assert _n in _v_and_c__or_n and v_and_c not in _v_and_c__or_n
        assert v not in _v_and_c__or_n and c not in _v_and_c__or_n
        assert v_and_n in _v_and_n__or_n and _n in _v_and_n__or_n
        assert v not in _v_and_n__or_n and n not in _v_and_n__or_n

    if disp:
        print("v_and_c__or_n:%s" % _v_and_c__or_n)
        print("v_and_c__or_c:%s" % _v_and_c__or_c)
        print("v_and_n__or_n:%s" % _v_and_n__or_n)

    if disp: print("individual conditions...")
    __v = _v(s, info)
    __c = _c(s, info)
    __n = _n(s, info)

    assert __v and __n
    assert not __c
    if info == 'self':
        assert v in __v and n in __n

    if disp:
        print("v:%s" % __v)
        print("c:%s" % __c)
        print("n:%s" % __n)

    if disp: print("conditions with false first...")
    _c_and_v = c_and_v(s, info)
    _c_or_v = c_or_v(s, info)
    _c_or__c_and_v = c_or__c_and_v(s, info)

    assert not _c_and_v
    assert _c_or_v
    assert not _c_or__c_and_v
    if info == 'self':
        assert v in _c_or_v and c not in _c_or_v

    if disp:
        print("c_and_v:%s" % _c_and_v)
        print("c_or_v:%s" % _c_or_v)
        print("c_or__c_and_v:%s" % _c_or__c_and_v)


if __name__ == '__main__':
    test_me(False)
    test_me(True)
    test_me()


# EOF



#!/usr/bin/env python
#
# Problem definition:
# Example in google/or-tools
# https://github.com/google/or-tools/blob/master/examples/python/integer_programming.py
# with Copyright 2010-2013 Google
# and disclamer as stated at the above reference link.
# 
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
  Integer programming example

  Minimize:
            f(x) = x0 + 2*x1

  Where: 
            3*x0 + 2*x1 >= 17
            x0,x1 are integers
"""

def objective(x):
    return x[0] + 2*x[1]

from mystic.bounds import Bounds
bounds = Bounds(0,None,n=2)

# with penalty='penalty' applied, solution is:
xs = [6., 0.]
ys = 6.0

# constraints
equations = """
3.*x0 + 2.*x1 - 17. >= 0.0
"""

from mystic.symbolic import generate_penalty, generate_conditions
pf = generate_penalty(generate_conditions(equations))
from mystic.symbolic import generate_constraint, generate_solvers, simplify
cf = generate_constraint(generate_solvers(simplify(equations)))

from mystic.constraints import integers

@integers()
def round(x):
  return x


if __name__ == '__main__':

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    result = diffev2(objective, x0=bounds, bounds=bounds, penalty=pf, constraints=round, npop=20, gtol=50, disp=True, full_output=True)

    print(result[0])
    assert almostEqual(result[0], xs, tol=1e-8) #XXX: fails b/c rel & zero?
    assert almostEqual(result[1], ys, tol=1e-4)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Alta Fang (altafang @caltech and alta @princeton)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = _doc = """
This is drawn from examples in Applied Optimization with MATLAB programming,
with the function definition found in [1].

References:
    1. Venkataraman, P. "Applied Optimization with MATLAB Programming",
       John Wiley and Sons, Hoboken NJ, 2nd Edition, 2009.
"""
__all__ = ['Sinc', 'venkat91']

from .abstract_model import AbstractFunction

from math import sin

class Sinc(AbstractFunction):
    __doc__ = \
    """a Venkataraman's sinc function generator

Venkataraman's sinc function [1] has the global minimum at the center
of concentric rings of local minima, with well depth decreasing with
distance from center.

The generated function f(x) is identical to equation (9.5) of example
9.1 of [1], and requires len(x) == 2.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates Venkataraman's sinc function for a list of coeffs

f(x) = -20 * \sin(r(x))/r(x)

Where:
r(x) = \sqrt((x_0 - 4)^2 + (x_1 - 4)^2 + 0.1)

Inspect with mystic_model_plotter using::
    mystic.models.venkat91 -b "-10:10:.1, -10:10:.1" -d

The minimum is f(x)=-19.668329370585823 at x=(4.0, 4.0)"""
        x,y = coeffs
        R = ((x-4.)**2 + (y-4.)**2 + 0.1)**.5
        return -20. * sin(R)/R

    minimizers = None #XXX: there are many local minima
    pass

# cleanup
del _doc

# prepared instances
venkat91 = Sinc().function

# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Lorentzian peak model

References:
    None
"""
__all__ = ['Lorentzian', 'lorentzian', 'gendata', 'histogram']

from .abstract_model import AbstractModel

from numpy import sum as numpysum
from numpy import array, pi, asarray, arange
import random

class Lorentzian(AbstractModel):
    """
Computes lorentzian
    """

    def __init__(self,name='lorentz',metric=lambda x: numpysum(x*x),sigma=1.0):
        AbstractModel.__init__(self,name,metric,sigma)
        return

    def evaluate(self,coeffs,evalpts):
        """evaluate lorentzian with given coeffs over given evalpts
coeffs = (a1,a2,a3,A0,E0,G0,n)"""
        a1,a2,a3,A0,E0,G0,n = coeffs
        x = asarray(evalpts) #XXX: requires a numpy.array
        return (a1 + a2*x + a3*x*x + A0 * ( G0/(2*pi) )/( (x-E0)*(x-E0)+(G0/2)*(G0/2) ))/n

    def ForwardFactory(self,coeffs):
        """generates a lorentzian model instance from a list of coefficients"""
        a1,a2,a3,A0,E0,G0,n = coeffs
        def forward_lorentzian(evalpts):
            """a lorentzian peak over a 1D numpy array
with (a1,a2,a3,A0,E0,G0,n) = (%s,%s,%s,%s,%s,%s,%s)""" % (a1,a2,a3,A0,E0,G0,n)
            return self.evaluate((a1,a2,a3,A0,E0,G0,n),evalpts)
        return forward_lorentzian

    pass
 

# prepared instances
lorentzian = Lorentzian()

def gendata(params, xmin, xmax, npts=4000):
    """Generate a lorentzian dataset of npts between [min,max] from given params"""
    F = lorentzian.ForwardFactory
    def gensample(F, xmin, xmax):
        from numpy import arange
        import random
        a = arange(xmin, xmax, (xmax-xmin)/200.)
        ymin = 0
        ymax = F(a).max()
        while 1:
            t1 = random.random() * (xmax-xmin) + xmin
            t2 = random.random() * (ymax-ymin) + ymin
            t3 = F(t1)
            if t2 < t3:
                return t1
    fwd = F(params)
    return array([gensample(fwd, xmin,xmax) for i in range(npts)])

# probably shouldn't be in here...
from numpy import histogram as numpyhisto
def histogram(data, binwidth, xmin, xmax):
    """generate bin-centered histogram of provided data
return bins of given binwidth (and histogram) generated between [xmin,xmax]"""
    bins = arange(xmin,xmax, binwidth)
    binsc = bins + (0.5 * binwidth)
    try: #FIXME: use updated numpy.histogram
        histo = numpyhisto(data, bins, new=False)[0]
    except:
        histo = numpyhisto(data, bins)[0]
    return binsc[:len(histo)], histo


# End of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2024-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
test rbf interpolation
"""

import numpy as np
x = np.random.rand(1,3)

def cost(x):
   return sum(np.sin(x)**2) + 1

y = cost(x.T)

import mystic.math.interpolate as ip

# functions which are non-zero at r = 0 reproduce a single datapoint
fx = ip.Rbf(*np.vstack((x.T, y)), function='gaussian', smooth=0.0)
assert (fx(*x.T) - y).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='multiquadric', smooth=0.0)
assert (fx(*x.T) - y).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='inverse', smooth=0.0)
assert (fx(*x.T) - y).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='inverse_quadratic', smooth=0.0)
assert (fx(*x.T) - y).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='bump', smooth=0.0)
assert np.abs(fx(*x.T) - y).sum() <= 1e-15 #NOTE: bump can mildly fail

# functions which are zero at r = 0 are singular with a single datapoint
fx = ip.Rbf(*np.vstack((x.T, y)), function='linear', smooth=-1e-100)
assert fx(*x.T).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='cubic', smooth=-1e-100)
assert fx(*x.T).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='quintic', smooth=-1e-100)
assert fx(*x.T).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='thin_plate', smooth=-1e-100)
assert fx(*x.T).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='quartic', smooth=-1e-100)
assert fx(*x.T).sum() == 0.0
fx = ip.Rbf(*np.vstack((x.T, y)), function='hyperbolic', smooth=-1e-100)
assert fx(*x.T).sum() == 0.0



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
In test_twistedgaussian, we compared SCEM (one chain) with metropolis.

Now we will use n-chain SCEM .
"""

from test_twistedgaussian import *

a = initpop(q*m, n)
b = list(map(target, a))
a,b = sort_ab_with_b(a,b)
Cs = sequential_deal(a, q)
As = [xx.tolist() for xx in sequential_deal(b, q)]

if __name__=='__main__':
    from mystic.metropolis import *
    import time

    Sk = [ [Cs[i][0]] for i in range(q) ]
    Sak = [ [As[i][0]] for i in range(q) ]

    for iter in range(5):
       # this is parallel
       print("iteration: %s" % str(iter+1))
       for chain in range(q):
           for i in range(1000):
              scem(Cs[chain], As[chain], Sk[chain], Sak[chain], target, 0.1)

       
       # need to gather and remix
       Cs , As = remix(Cs, As)
    
    from mystic.tools import flatten_array

    Sk = [a[100:] for a in Sk] # throw away the first 100 pts of each chain
    sk = flatten_array(Sk,1)
    #print("length of sk: %s" % len(sk))
    
    import matplotlib.pyplot as plt
    plt.plot(sk[:,0],sk[:,1],'r.')
    plt.show()

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2022-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
print the size of the database of sampled data from xrd_design*.py
"""

import os
import numpy as np
import mystic.cache.archive as mca
archives = ('ave.db', 'min.db', 'max.db')

for archive in archives:
    a = None
    if os.path.exists(archive):
        try:
            a = mca.read(archive, type=mca.file_archive)
            print('len(%s): %s' % (archive, len(a)))
            #a = np.array(list(a.values()))
            #print('  min=%s, ave=%s, max=%s' % (a.min(), a.mean(), a.max()))
        except Exception:
            pass



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Minimize Rosenbrock's Function with Nelder-Mead.
    - Plot of parameter convergence to function minimum.

Demonstrates:
    - standard models
    - minimal solver interface
    - parameter trajectories using retall
"""

# Nelder-Mead solver
from mystic.solvers import fmin

# Rosenbrock function
from mystic.models import rosen

# tools
import matplotlib.pyplot as plt


if __name__ == '__main__':

    # initial guess
    x0 = [0.8,1.2,0.7]

    # use Nelder-Mead to minimize the Rosenbrock function
    solution = fmin(rosen,x0,disp=0,retall=1)
    allvecs = solution[-1]

    # plot the parameter trajectories
    plt.plot([i[0] for i in allvecs])
    plt.plot([i[1] for i in allvecs])
    plt.plot([i[2] for i in allvecs])

    # draw the plot
    plt.title("Rosenbrock parameter convergence")
    plt.xlabel("Nelder-Mead solver iterations")
    plt.ylabel("parameter value")
    plt.legend(["x", "y", "z"])
    plt.show()
 
# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
solvers: minimal and expanded interfaces for optimization algorithms


Standard Interface
==================

All of mystic's optimizers derive from the solver API, which provides
each optimizer with a standard, but highly-customizable interface.
A description of the solver API is found in ``mystic.models.abstract_model``,
and in each derived optimizer.  Mystic's optimizers are::

    ** Global Optimizers **
    DifferentialEvolutionSolver  -- Differential Evolution algorithm
    DifferentialEvolutionSolver2 -- Price & Storn's Differential Evolution
    ** Pseudo-Global Optimizers **
    BuckshotSolver               -- Uniform Random Distribution of N Solvers
    LatticeSolver                -- Distribution of N Solvers on a Regular Grid
    SparsitySolver               -- N Solvers Sampled where Point Density is Low
    ResidualSolver               -- N Solvers Sampled where Model Misfit is High
    MixedSolver                  -- Mixture of N Pseudo-Global Optimizers
    ** Local-Search Optimizers **
    NelderMeadSimplexSolver      -- Nelder-Mead Simplex algorithm
    PowellDirectionalSolver      -- Powell's (modified) Level Set algorithm


Minimal Interface
=================

Most of mystic's optimizers can be called from a minimal (i.e. one-line)
interface. The collection of arguments is often unique to the optimizer,
and if the underlying solver derives from a third-party package, the
original interface is reproduced. Minimal interfaces to these optimizers
are provided::

    ** Global Optimizers **
    diffev      -- DifferentialEvolutionSolver
    diffev2     -- DifferentialEvolutionSolver2
    ** Pseudo-Global Optimizers **
    buckshot    -- BuckshotSolver
    lattice     -- LatticeSolver
    sparsity    -- SparsitySolver
    residual    -- ResidualSolver
    ** Local-Search Optimizers **
    fmin        -- NelderMeadSimplexSolver
    fmin_powell -- PowellDirectionalSolver


More Information
================

For more information, please see the solver documentation found here:
    - ``mystic.differential_evolution``   [differential evolution solvers]
    - ``mystic.scipy_optimize``           [scipy local-search solvers]
    - ``mystic.ensemble``                 [pseudo-global solvers]

or the API documentation found here:
    - ``mystic.abstract_solver``          [the solver API definition]
    - ``mystic.abstract_map_solver``      [the parallel solver API]
    - ``mystic.abstract_ensemble_solver`` [the ensemble solver API]

"""
# global optimizers
from mystic.differential_evolution import DifferentialEvolutionSolver
from mystic.differential_evolution import DifferentialEvolutionSolver2
from mystic.differential_evolution import diffev, diffev2

# pseudo-global optimizers
from mystic.ensemble import BuckshotSolver
from mystic.ensemble import LatticeSolver
from mystic.ensemble import SparsitySolver
from mystic.ensemble import ResidualSolver
from mystic.ensemble import MixedSolver
from mystic.ensemble import buckshot, lattice, sparsity, residual

# local-search optimizers
from mystic.scipy_optimize import NelderMeadSimplexSolver
from mystic.scipy_optimize import PowellDirectionalSolver
from mystic.scipy_optimize import fmin, fmin_powell


# load a solver from a restart file
def LoadSolver(filename=None, **kwds):
    """load solver state from a restart file

Args:
    filename (str, default=None): path of solver restart file
    **kwds (dict, default={}): solver state overrides

Returns:
    solver instance
    """
    if filename is None:
        filename = kwds['_state'] if '_state' in kwds else None
    #XXX: only allow a list override keys (lookup values from self)
#   if filename is None: filename = self._state
#   if filename is None:
#       solver = self
#   else:
    import dill
    if filename: f = open(filename, 'rb')
    else: return
    try:
        solver = dill.load(f)
        _locals = {}
        _locals['solver'] = solver
        code = "from mystic.solvers import %s;" % solver._type
        code += "self = %s(solver.nDim);" % solver._type
        code = compile(code, '<string>', 'exec')
        exec(code, _locals)
        self = _locals['self']
    finally:
        f.close()
    # transfer state from solver to self, allowing overrides
    self._AbstractSolver__load_state(solver, **kwds)
    self._state = filename
    self._stepmon.info('LOADED("%s")' % filename)
    return self


# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from mystic.constraints import *

def _symbolic():
    import mystic.symbolic as ms
    eqn = """x0 - 2*x1 > 0
             x3 - 3*x2 < 0"""
    eqn = ms.simplify(eqn, target=['x1','x3'])
    cons = ms.generate_constraint(ms.generate_solvers(eqn))
    return cons

def _mean():
    import mystic.constraints as mc
    return mc.with_mean(5.0)(lambda x: x)

def _parameter():
    def constraint(x):
        x[-1] = x[0]
        return x
    return constraint

def _compound():
    import mystic.constraints as mc
    return mc.or_(_parameter(), _mean(), _symbolic())

def test_vectorize(constraint, axis):
    # generate data
    import numpy as np
    data = np.random.randn(10,4)

    # build transform
    ineq = vectorize(constraint, axis)

    # test transform
    res = ineq(data)
    assert np.all(ineq(res) == res)

def test_vectorize_sklearn(constraint, axis):
    # get dataset
    from sklearn.datasets import load_iris
    iris = load_iris()

    # build transform
    ineq = vectorize(constraint, axis)
    from sklearn.preprocessing import FunctionTransformer
    t = FunctionTransformer(func=ineq, validate=False) #XXX: inverse?

    # test transform
    import numpy as np
    iris_ = t.fit(iris.data).transform(iris.data)
    assert np.all(t._transform(iris_) == iris_)


if __name__ == '__main__':
    try:
        import sklearn
        HAS_SKLEARN = True
    except ImportError:
        HAS_SKLEARN = False

    for cons in (_symbolic, _mean, _parameter, _compound):
        for axis in (0,1):
            test_vectorize(cons(), axis)
            if HAS_SKLEARN: test_vectorize_sklearn(cons(), axis)

# EOF



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

def objective(x):
    x0,x1,x2,x3 = x
    return 3*x0 + 1.e-6*x0**3 + 2*x1 + 2.e-6/3*x1**3

from mystic.bounds import Bounds
bounds = Bounds([0,-0.55],[1200,0.55], n=(2,2))
# with penalty='penalty' applied, solution is:
xs = [679.945794311, 1026.0666256385, 0.11887602615356, -0.3962337137961]
ys = 5126.49810960

from mystic.symbolic import generate_constraint, generate_solvers, simplify
from mystic.symbolic import generate_penalty, generate_conditions

equations = """
x2 - x3 - 0.55 <= 0.0
x3 - x2 - 0.55 <= 0.0
abs(1000*(sin(-x2-.25) + sin(-x3-0.25)) + 894.8 - x0) = 0.0
abs(1000*(sin(x2-.25) + sin(x2-x3-0.25)) + 894.8 - x1) = 0.0
abs(1000*(sin(x3-.25) + sin(x3-x2-0.25)) + 1294.8) = 0.0
"""
#cf = generate_constraint(generate_solvers(simplify(equations)))
pf = generate_penalty(generate_conditions(equations), k=1e12)



if __name__ == '__main__':

    from mystic.solvers import buckshot, sparsity
    from mystic.math import almostEqual

    result = buckshot(objective, len(xs), npts=100, bounds=bounds, penalty=pf, disp=False, full_output=True)
    #result = sparsity(objective, len(xs), npts=100, rtol=-10, bounds=bounds, penalty=pf, disp=False, full_output=True)

    assert almostEqual(result[0], xs, rel=1e-1)
    assert almostEqual(result[1], ys, rel=1e-1)



# EOF



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Implements the "Shuffled Complex Evolution Metropolis" Algoritm
of Vrugt et al. [1]

  
Reference:

[1] Jasper A. Vrugt, Hoshin V. Gupta, Willem Bouten, and Soroosh Sorooshian
A Shuffled Complex Evolution Metropolis algorithm for optimization and
uncertainty assessment of hydrologic model parameters,
WATER RESOURCES RESEARCH, VOL. 39, NO. 8, 1201, doi:10.1029/2002WR001642, 2003 
http://www.agu.org/pubs/crossref/2003/2002WR001642.shtml

[2] Vrugt JA, Nuallain , Robinson BA, Bouten W, Dekker SC, Sloot PM
Application of parallel computing to stochastic parameter estimation in
environmental models,
Computers & Geosciences, Vol. 32, No. 8. (October 2006), pp. 1139-1155.
http://www.science.uva.nl/research/scs/papers/archive/Vrugt2006b.pdf

"""

import numpy
try:
    getlimits = numpy.lib.getlimits
except AttributeError:
    try:
        getlimits = numpy._core.getlimits
    except AttributeError:
        getlimits = numpy.core.getlimits

TYPE_DOUBLE =  getlimits.finfo(getlimits.numeric.double)
TINY = TYPE_DOUBLE.tiny

def multinormal_pdf(mean, var):
    """var must be symmetric positive definite """
    import numpy
    vinv = numpy.linalg.inv(var)
    mu = numpy.array(mean).flatten()
    n = len(mu)

    # check that var is properly sized
    dum = numpy.dot(mu,var) - numpy.dot(var,mu)

    prefactor = 1./numpy.sqrt((2 * numpy.pi)**n  * numpy.linalg.det(var))
    def _(x):
        xm = numpy.array(x) - mu
        return prefactor *  numpy.exp(-0.5 * numpy.dot(xm,numpy.dot(vinv,xm)) )
    return _

def sequential_deal(inarray, n):
    """
- inarray: should be a set of N objects (the objects can be vectors themselves, 
  but inarray should be index-able like a list. It is coerced into a numpy array
  because the last operations requires that it is also indexable by a 'list.'
    
- it should have a length divisble by n, otherwise the reshape will fail 
  (this is a feature !) 

- sequential_deal(range(20), 5) wil return a 5 element list, each element being
  a 4-list of index. (see below)

>>> for l in sequential_deal(range(20),5):
...    print(l)
...
[ 0  5 10 15]
[ 1  6 11 16]
[ 2  7 12 17]
[ 3  8 13 18]
[ 4  9 14 19]

    """
    import numpy
    cards = numpy.array(inarray)
    N = len(cards)
    # this bit of numpy will give, for N=20, n = 5
    # ord = [ [0,5,10,15], [1,6,11,16], [2,7,12,17], [3,8,13,18], [4,9,14,19] ]
    ord = numpy.transpose(numpy.array(list(range(N))).reshape(N//n, n))
    return [cards[x] for x in ord] 

def sort_and_deal(cards, target, nplayers):
    import numpy
    c = numpy.array(list(map(target, cards)))
    o = list(reversed(numpy.argsort(c)))
    # from best to worst
    sorted_deck = cards[o]
    return sequential_deal(sorted_deck, nplayers)

def sort_ab_with_b(a, b, ord = -1):
    """default is descending..."""
    import numpy
    aa,bb = numpy.array(a), numpy.array(b)
    if ord == -1:
        o = list(reversed(bb.argsort()))
    else:
        o = b.argsort()
    return aa[o], bb[o]

def sort_complex0(c, a):
    # this is dumb, because c (i.e., a, are almost sorted)
    # should use the one below instead.
    # this is faster than sort_complex
    import numpy
    b = numpy.array(a)
    o = list(reversed(b.argsort()))
    return c[o], list(b[o])

def sort_complex(c, a):
    # this is dumb, because c (i.e., a, are almost sorted)
    # should use the one below instead.
    import numpy
    D = list(zip(a,c))
    D.sort(key=lambda x:x[0], reverse=True)
    return numpy.array([x[1] for x in D]), [x[0] for x in D]

def myinsert(a, x):
    # a is in descending order
    from bisect import bisect
    return bisect([-i for i in a],-x)

def sort_complex2(c, a):
    """
- c and a are partially sorted (either the last one is bad, or the first one)
- pos : 0 (first one out of order)
       -1 (last one out of order)
    """
    # find where new key is relative to the rest
    if a[0] < a[1]:
        ax, cx = a[0], c[0]
        ind = myinsert(a[1:],ax)
        a[0:ind], a[ind] = a[1:1+ind], ax
        c[0:ind], c[ind] = c[1:1+ind], cx
    elif a[-1] > a[-2]:
        ax,cx = a[-1], c[-1]
        ind = myinsert(a[0:-1],ax)
        a[ind+1:], a[ind] = a[ind:-1], ax
        c[ind+1:], c[ind] = c[ind:-1], cx
    return

def update_complex(Ck, ak, c, a, pos):
    """
- ak is sorted (descending)
- Ck[pos] and ak[pos] will be removed, and then
  c and a spliced in at the proper place
- pos is 0, or -1
    """
    if pos == 0:
        at = ak[1:]
    else:
        at = ak[:-1]

    p = myinsert(at, a)

    # not done yet.
    if pos == 0:
        Ck[0:p], ak[p] = Ck[1:p+1], ak[1:p+1]
        Ck[p], ak[p] = c, a
    else:
        Ck[0:p], ak[p] = Ck[1:p+1], ak[1:p+1]
        Ck[p], ak[p] = c, a

    return

def scem(Ck, ak, Sk, Sak, target, cn):
    """
This is the SCEM algorithm starting from line [35] of the reference [1].

- [inout] Ck is the kth 'complex' with m points. This should be an m by n array
  n being the dimensionality of the density function. i.e., the data
  are arranged in rows. 

  Ck is assumed to be sorted according to the target density.

- [inout] ak, the density of the points of Ck.

- [inout] Sk, the entire chain. (should be a list)

- [inout] Sak, the cost of the entire chain (should be a list)

  Sak would be more convenient to use if it is a numpy array, but we need
  to append to it frequently.

- [in] target: target density function

- [in] cn: jumprate. (see Paragraph 37 of [1.]

- The invariants: ak is always aligned with Ck, and are the cost of Ck
- Similarly, Sak is always aligned with Sk in the same way.

- On return... sort order in Ck/ak is destroyed. but see sort_complex2

    """
    import numpy
    from mystic.tools import random_state
    prng = random_state(module=numpy.random)

    # function level constants
    T = 100000. # predefined likelihood ratio. Paragraph 45 of [1.]

    # Sort Ck according to ak
    #Ck, ak = sort_complex0(Ck, ak) 
    #print("ak before: %s" % ak)
    #Ck, ak = sort_complex(Ck, ak) 
    sort_complex2(Ck, ak) 
    #print("ak after: %s" % ak)

    # number of points per complex
    M = Ck.shape[0]

    mu = numpy.mean(Ck,0) # row mean

    # (numpy.cov takes data in columns)
    Sigma = numpy.cov(numpy.transpose(Ck)) 

    # Gamma (line 35 of [1]. Best to Worst)
    Gamma = ak[0] / (ak[-1]+TINY)
   
    if len(Sak) >= M:
        meansak = numpy.mean(Sak[-M:])
    else:
        meansak = numpy.mean(Sak)
    alpha_k = numpy.mean(ak) / (meansak+TINY)
  
    if alpha_k < T:
         # Paragraph 37 of [1]
         basept = Sk[-1]
    else: # numpy.mean(asak) is very close to zero.
         # Paragraph 38 of [1]
         basept = mu

    # TODO should take a proposal instead !
    Yt = prng.multivariate_normal(basept, cn*cn * Sigma)
    cY = target(Yt)    

    # print("new/orig : %s %s" % (cY, Sak[-1]))

    r = min( cY / (Sak[-1]+TINY), 1)

    if prng.rand() <= r:
        Sk.append(Yt)
        Sak.append(cY)
        # Paragraph 43 of [1] (update the best of Ck)
        Ck[0], ak[0] = Sk[-1], Sak[-1]
    else:
        Sk.append(Sk[-1])
        Sak.append(Sak[-1])
        if Gamma > T and Sak[-1] > ak[-1]:
            # Paragraph 43 of [1] (update the worst of Ck)
            Ck[-1], ak[-1] = Sk[-1], Sak[-1]

    # mainly because of the logic in Paragraph [43] that updates the complex,
    # this function is not "functional" and modifies its argument instead.
    # hence no return value
    return     

def remix(Cs, As):
    """
Mixing and dealing the complexes.
The types of Cs and As are very important.... 
    """
    from mystic.tools import flatten_array
    q = len(Cs)
    c2, a2 =  list(flatten_array(Cs,1)), list(flatten_array(As,1))
    a,b = sort_ab_with_b(c2,a2)
    C2 = sequential_deal(a, q)
    A2 = [xx.tolist() for xx in sequential_deal(b, q)]
    return C2, A2

# end of file



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2009-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

"""Original matlab code:

function A=marc_surr(x)
h=x(1)*25.4*10^(-3);
a=x(2)*pi/180;
v=x(3);
Ho=0.5794;
s=1.4004;
n=0.4482;
K=10.3963;
p=0.4757;
u=1.0275;
m=0.4682;
Dp=1.778;

v_bl=Ho*(h/(cos(a))^(n))^s;

if v<v_bl
  A=0
else
  A=K*((h/Dp)^p)*((cos(a))^u)*(tanh((v/v_bl)-1))^m;
end
"""

### NOTES ###
# h = thickness = [60,105]
# a = obliquity = [0,30]
# v = speed = [2.1,2.8]

# explore the cuboid (h,a,v), with
# subdivisions at h=100, a=20, v=2.2
# due to ballistic limit: v(h=100,a=20) = 2.22,
# perforation in this region should be zero.
# NOTE: 'failure' is A < = t
# 
# Calculate for each of the 8 subcuboids:
#  * probability mass, i.e. the product of the normalized side-lengths,
#    since we're taking h, a and v to be uniformly distributed in their
#    intervals
#  * McDiarmid diameter of the perforation area A when restricted to that
#    cuboid or subcuboid
#  * the mean value of the perforation area A on each (sub)cuboid

from math import pi, cos, tanh

def ballistic_limit(h,a):
  """calculate ballistic limit

  Inputs:
    - h = thickness in (unknown) units
    - a = obliquity in (unknown) units

  Outputs:
    - v_bl = velocity (ballistic limit) in (unknown) units
"""
 #h = x[0] * 25.4 * 1e-3
 #a = x[1] * pi/180.0
  Ho = 0.5794
  s = 1.4004
  n = 0.4482
  return Ho * ( h / cos(a)**n )**s


def marc_surr(x):
  """calculate perforation area using a tanh-based model surrogate

  Inputs:
    - x = [thickness, obliquity, speed] in (unknown) units

  Outputs:
    - A = performation area in (unknown) units
"""
# h = thickness = [60,105]
# a = obliquity = [0,30]
# v = speed = [2.1,2.8]
  h = x[0] * 25.4 * 1e-3
  a = x[1] * pi/180.0
  v = x[2]

  K = 10.3963
  p = 0.4757
  u = 1.0275
  m = 0.4682
  Dp = 1.778

  # compare to ballistic limit
  v_bl = ballistic_limit(h,a)
  if v < v_bl:
    return 0

  return K * (h/Dp)**p * (cos(a))**u * (tanh((v/v_bl)-1))**m

# EOF



#!/usr/bin/env python
#
# Author: Patrick Hung (patrickh @caltech)
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Solve the dual form of test_circle.py.

Given a set of points in the plane, find the smallest circle
that contains them.

Requires:
  -- numpy, matplotlib

The matplotlib output will draw 
  -- a set of points inside a circle defined by x0,y0,R0 
  -- the circle (x0,y0) with rad R0
  -- the optimized circle with minimum R enclosing the points
"""

from numpy import *
import matplotlib.pyplot as plt
from test_circle import sparse_circle, circle, sv, x0, y0, R0

# a common objective function for solving a QP problem
# (see http://www.mathworks.com/help/optim/ug/quadprog.html)
def objective(x, H, f):
    return 0.5 * dot(dot(x,H),x) + dot(f,x)

# plot the data and results
def plot(xy, sv, x0, y0, R0, center, R):
    import matplotlib.pyplot as plt
    plt.plot(xy[:,0],xy[:,1],'k+')
    plt.plot(xy[sv,0],xy[sv,1],'bx',mfc="None")
    plt.plot([x0],[y0],'ro')
    plt.plot([center[0]],[center[1]],'bo')
    plt.plot([xy[sv[0],0], center[0]],[xy[sv[0],1], center[1]],'b--')
    c = circle(x0,y0,R0)
    plt.plot(c[:,0], c[:,1], 'r-', linewidth=2)
    c = circle(center[0],center[1],R)
    plt.plot(c[:,0], c[:,1], 'b-', linewidth=2)
    plt.axis('equal')
    plt.show()


if __name__ == '__main__':
    npt = 20
    from test_circle import xy
    npt1 = xy.shape[0]
    if npt is not npt1:
        xy = sparse_circle(x0,y0,R0,npt)
    else:
        pass

    # define a QP problem for the objective
    Q = dot(xy, transpose(xy))
    f = -diag(Q)+10
    H = Q*2
    # define lower and upper bounds on x
    LB = zeros(npt)
    UB = ones(npt)
    # define equality constraints (A*x == b)
    A = ones((1,npt))
    b = ones(1)

#   # generic: build a constraint where (A*x == b)
#   from mystic.symbolic import linear_symbolic, solve, \
#        generate_solvers as solvers, generate_constraint as constraint
#   norm = constraint(solvers(solve(linear_symbolic(A,b),target=['x0'])))

    # specific: build a constraint where (sum(x) == 1.0)
    from mystic.math.measures import normalize
    def norm(x):
        return normalize(x, mass=1.0)

    # solve the constrained quadratic programming problem
    from mystic.solvers import fmin_powell
    x = fmin_powell(objective, UB/npt, args=(H, f), bounds=list(zip(LB,UB)), \
                    constraints=norm, ftol=1e-8, disp=1)

    # find support vectors
    sv = where(x > 0.001)[0]
    print("support vectors: %s" % sv)

    # compare solved center and radius to generating center and radius
    center = dot(x,xy)
    R = linalg.norm(xy[sv[0],:]-center)
    print("x0, y0: (%f, %f) @ R0 = %f" % (x0,y0,R0))
    print("center: (%f, %f) @ R  = %f" % (center[0],center[1],R))

    plot(xy, sv, x0, y0, R0, center, R)


# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2020-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
'''
demonstrate iterative improvement of R^2 score for an ANN model
'''
from ml import *


if __name__ == '__main__':

    # get model data
    from sklearn.datasets import load_iris
    data = load_iris()
    d = MLData(*traintest(data.data[:,:3], data.data[:,3], .2))

    # build estimator and perform initial fit
    from sklearn.preprocessing import StandardScaler
    from sklearn.neural_network import MLPRegressor
    xfm = StandardScaler().fit(d.xtrain)
    mlp = MLPRegressor().fit(xfm.transform(d.xtrain), d.ytrain)
    e = Estimator(mlp, xfm)

    # check the initial score
    print('initial score')
    print(e.score(d.xtest, d.ytest))

    # iteratively improve the score
    print('improving...')
    ee = improve_score(e, d, verbose=True)
    print('checking improved score')
    print(ee.score(d.xtest, d.ytest))
    print('improving...')
    ee = improve_score(ee, d, verbose=True)
    print('checking improved score')
    print(ee.score(d.xtest, d.ytest))




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
#
# Adapted from parallel_desolve.py

__doc__ = """
# Tests MPI version of Storn and Price's Polynomial 'Fitting' Problem,
# using a single solver with the population launched in parallel.
# This can be much slower, due to overhead of repeatedly setting up MPI.
# 
# Exact answer: [1,1,1]
  
# Reference:
#
# [1] Storn, R. and Price, K. Differential Evolution - A Simple and Efficient
# Heuristic for Global Optimization over Continuous Spaces. Journal of Global
# Optimization 11: 341-359, 1997.

# To run in parallel:  (must install 'pyina')
python ezmap_desolve_rosen.py
"""

try:
  from pyina.launchers import Mpi as Pool
# from pyina.launchers import TorqueMpi as Pool
except:
  print(__doc__)


from mystic.solvers import DifferentialEvolutionSolver2
from mystic.termination import ChangeOverGeneration, VTR
from mystic.strategy import Best1Exp
from mystic.monitors import VerboseMonitor
from mystic.tools import random_seed

#from raw_rosen import rosen as myCost     # with a helper function
from mystic.models import rosen as myCost  # without a helper function

ND = 3
NP = 20
MAX_GENERATIONS = NP*NP
NNODES = "5:ppn=4"
QUEUE = "weekdayQ"
TIMELIMIT = "00:30:00"

TOL = 0.01
CROSS = 0.9
SCALE = 0.8
seed = 100


if __name__=='__main__':
    def print_solution(func):
        print(func)
        return

    psow = VerboseMonitor(10)
    ssow = VerboseMonitor(10)

    random_seed(seed)
    print("first sequential...")
    solver = DifferentialEvolutionSolver2(ND,NP)  #XXX: sequential
    solver.SetRandomInitialPoints(min=[-100.0]*ND, max=[100.0]*ND)
    solver.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver.SetGenerationMonitor(ssow)
    solver.Solve(myCost, VTR(TOL), strategy=Best1Exp, \
                 CrossProbability=CROSS, ScalingFactor=SCALE, disp=1)
    print("")
    print_solution( solver.bestSolution )

    random_seed(seed)
    print("\n and now parallel...")
    solver2 = DifferentialEvolutionSolver2(ND,NP)  #XXX: parallel
    solver2.SetMapper(Pool(NNODES).map)
#   solver2.SetMapper(Pool(NNODES, queue=QUEUE, timelimit=TIMELIMIT).map)
    solver2.SetRandomInitialPoints(min=[-100.0]*ND, max=[100.0]*ND)
    solver2.SetEvaluationLimits(generations=MAX_GENERATIONS)
    solver2.SetGenerationMonitor(psow)
    solver2.Solve(myCost, VTR(TOL), strategy=Best1Exp, \
                  CrossProbability=CROSS, ScalingFactor=SCALE, disp=1)
    print("")
    print_solution( solver2.bestSolution )

# end of file



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE

from g05 import objective, bounds, xs, ys

from mystic.constraints import as_constraint
from mystic.penalty import quadratic_inequality, quadratic_equality

def penalty1(x): # <= 0.0
    return x[2] - x[3] - 0.55

def penalty2(x): # <= 0.0
    return x[3] - x[2] - 0.55

def penalty3(x): # == 0.0
    from math import sin
    return abs(1000*(sin(-x[2]-.25) + sin(-x[3]-0.25)) + 894.8 - x[0])

def penalty4(x): # == 0.0
    from math import sin
    return abs(1000*(sin(x[2]-.25) + sin(x[2]-x[3]-0.25)) + 894.8 - x[1])

def penalty5(x): # == 0.0
    from math import sin
    return abs(1000*(sin(x[3]-.25) + sin(x[3]-x[2]-0.25)) + 1294.8)

@quadratic_inequality(penalty1, k=1e12)
@quadratic_inequality(penalty2, k=1e12)
@quadratic_equality(penalty3, k=1e12)
@quadratic_equality(penalty4, k=1e12)
@quadratic_equality(penalty5, k=1e12)
def penalty(x):
    return 0.0

solver = as_constraint(penalty)



if __name__ == '__main__':

    from mystic.solvers import buckshot, sparsity
    from mystic.math import almostEqual

    result = buckshot(objective, len(xs), npts=100, bounds=bounds, penalty=penalty, disp=False, full_output=True)
    #result = sparsity(objective, len(xs), npts=100, rtol=-10, bounds=bounds, penalty=penalty, disp=False, full_output=True)

    assert almostEqual(result[0], xs, rel=1e-1)
    assert almostEqual(result[1], ys, rel=1e-1)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @uqfoundation)
# Copyright (c) 2018-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
test _to_function and _to_objective in cost and gradient calculations
"""

import numpy as np
x = np.random.rand(10,3)

def cost(x):
   return sum(np.sin(x)**2) + 1

y = cost(x.T)

import mystic.math.interpolate as ip
fx = ip._to_function(cost)
assert (fx(*x.T) - cost(x.T) ).sum() < 0.0001

f = ip.interpf(x, y, method='linear', arrays=True)
cf = ip._to_objective(f)
assert (f(*x.T) - cf(x.T) ).sum() < 0.0001

assert f(*x[0].T) == cf(x[0].T)
assert fx(*x[0].T) == cost(x[0].T)

assert ip.gradient(x, f, method='linear').shape \
    == ip.gradient(x, fx, method='linear').shape


from mystic.models import rosen0der as rosen
y = rosen(x.T)

f = ip.interpf(x, y, method='linear')
fx = ip._to_function(rosen)
cf = ip._to_objective(f)

assert f(*x[0].T) == cf(x[0].T)
assert fx(*x[0].T) == rosen(x[0].T)
#print( fx(*x[0].T) - rosen(x[0].T) )

assert ip.gradient(x, f, method='linear').shape \
    == ip.gradient(x, fx, method='linear').shape




#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"""
Example:
    - Solve 8th-order Chebyshev polynomial coefficients with DE.
    - Callable plot of fitting to Chebyshev polynomial.
    - Monitor Chi-Squared for Chebyshev polynomial.
    - Impact of mutation and crossover coefficients

Demonstrates:
    - standard models
    - expanded solver interface
    - built-in random initial guess
    - solver interactivity
    - customized monitors and termination conditions
    - customized DE mutation strategies
    - use of monitor to retrieve results information
"""

# Differential Evolution solver
from mystic.solvers import DifferentialEvolutionSolver2

# Chebyshev polynomial and cost function
from mystic.models.poly import chebyshev8, chebyshev8cost
from mystic.models.poly import chebyshev8coeffs

# tools
from mystic.termination import VTR
from mystic.strategy import Best1Exp
from mystic.monitors import VerboseMonitor
from mystic.tools import getch, random_seed
from mystic.math import poly1d
import matplotlib.pyplot as plt
plt.ion()

# draw the plot
def plot_exact():
    plt.title("fitting 8th-order Chebyshev polynomial coefficients")
    plt.xlabel("x")
    plt.ylabel("f(x)")
    import numpy
    x = numpy.arange(-1.2, 1.2001, 0.01)
    exact = chebyshev8(x)
    plt.plot(x,exact,'b-')
    plt.legend(["Exact"])
    plt.axis([-1.4,1.4,-2,8])#,'k-')
    plt.draw()
    plt.pause(0.001)
    return
 
# plot the polynomial
def plot_solution(params,style='y-'):
    import numpy
    x = numpy.arange(-1.2, 1.2001, 0.01)
    f = poly1d(params)
    y = f(x)
    plt.plot(x,y,style)
    plt.legend(["Exact","Fitted"])
    plt.axis([-1.4,1.4,-2,8])#,'k-')
    plt.draw()
    plt.pause(0.001)
    return


if __name__ == '__main__':

    print("Differential Evolution")
    print("======================")

    # set range for random initial guess
    ndim = 9
    x0 = [(-100,100)]*ndim
    random_seed(123)

    # suggest that the user interacts with the solver
    print("NOTE: while solver is running, press 'Ctrl-C' in console window")
    getch()

    # draw frame and exact coefficients
    plot_exact()

    # configure monitor
    stepmon = VerboseMonitor(50)

    # use DE to solve 8th-order Chebyshev coefficients
    npop = 10*ndim
    solver = DifferentialEvolutionSolver2(ndim,npop)
    solver.SetRandomInitialPoints(min=[-100]*ndim, max=[100]*ndim)
    solver.SetEvaluationLimits(generations=999)
    solver.SetGenerationMonitor(stepmon)
    solver.enable_signal_handler()
    solver.Solve(chebyshev8cost, termination=VTR(0.01), strategy=Best1Exp, \
                 CrossProbability=0.8, ScalingFactor=0.5, \
                 sigint_callback=plot_solution)
    solution = solver.Solution()

    # use monitor to retrieve results information
    iterations = len(stepmon)
    cost = stepmon.y[-1]
    print("Generation %d has best Chi-Squared: %s" % (iterations, cost))

    # use pretty print for polynomials
    print(poly1d(solution))

    # compare solution with actual 8th-order Chebyshev coefficients
    print("\nActual Coefficients:\n %s\n" % poly1d(chebyshev8coeffs))

    # plot solution versus exact coefficients
    plot_solution(solution)
    getch()

# end of file



#!/usr/bin/env python
#
# Problem definition:
# A-R Hedar and M Fukushima, "Derivative-Free Filter Simulated Annealing
# Method for Constrained Continuous Global Optimization", Journal of
# Global Optimization, 35(4), 521-549 (2006).
# 
# Original Matlab code written by A. Hedar (Nov. 23, 2005)
# http://www-optima.amp.i.kyoto-u.ac.jp/member/student/hedar/Hedar_files/go.htm
# and ported to Python by Mike McKerns (December 2014)
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
"Welded Beam Design"

def objective(x):
    x0,x1,x2,x3 = x
    return 1.10471*x0**2*x1 + 0.04811*x2*x3*(14.0 + x1)

from mystic.bounds import Bounds
bounds = Bounds([0.125,.1],[10,10],n=(1,3))
# with penalty='penalty' applied, solution is:
xs = [ 0.20572964, 7.09241428, 9.03662391, 0.20572964]
ys = 2.21815086

# default parameters
P = 6000.
L = 14.
E = 30.e+6
G = 12.e+6
t_max = 13600.
s_max = 30000.
d_max = 0.25

# parameter equations
def M(x, P=P, L=L): return P*(L + x[1]/2.)
def R(x): return (0.25*(x[1]**2 + (x[0] + x[2])**2))**.5
def J(x): return 2./2**.5 * x[0] * x[1] * (x[1]**2/12. + 0.25*(x[0] + x[2])**2)
def P_c(x, L=L, E=E, G=G): return (4.013*E/(6*L**2)) * x[2]*x[3]**3 * (1-0.25*x[2]*(E/G)**.5/L)
def t1(x, P=P): return P/(2**.5 * x[0] * x[1])
def t2(x, P=P, L=L): return M(x,P,L)*R(x)/J(x)
def t(x, P=P, L=L): return (t1(x,P)**2 + t1(x,P)*t2(x,P,L)*x[1]/R(x) + t2(x,P,L)**2)**.5
def s(x, P=P, L=L): return 6*P*L/(x[3] * x[2]**2)
def d(x, P=P, L=L, E=E): return 4*P*L**3/(E * x[3] * x[2]**3)

from mystic.penalty import quadratic_inequality

def generate_penalty(**kwds):
    "enable override of P,L,E,G,t_max,s_max,d_max in penalties"

    def penalty1(x, P=P, L=L, t_max=t_max, **kwd): # <= 0.0
        return t(x,P,L) - t_max

    def penalty2(x, P=P, L=L, s_max=s_max, **kwd): # <= 0.0
        return s(x,P,L) - s_max

    def penalty3(x, **kwd): # <= 0.0
        return x[0] - x[3]

    def penalty4(x, **kwd): # <= 0.0
        return 0.10471*x[0]**2 + 0.04811*x[2]*x[3]*(14.0 + x[1]) - 5.0

    def penalty5(x, P=P, L=L, E=E, d_max=d_max, **kwd): # <= 0.0
        return d(x,P,L,E) - d_max

    def penalty6(x, P=P, L=L, E=E, G=G, **kwd): # <= 0.0
        return P - P_c(x,L,E,G)

    @quadratic_inequality(penalty1, k=1e12, kwds=kwds)
    @quadratic_inequality(penalty2, k=1e12, kwds=kwds)
    @quadratic_inequality(penalty3, k=1e12, kwds=kwds)
    @quadratic_inequality(penalty4, k=1e12, kwds=kwds)
    @quadratic_inequality(penalty5, k=1e12, kwds=kwds)
    @quadratic_inequality(penalty6, k=1e12, kwds=kwds)
    def penalty(x):
        return 0.0

    return penalty



if __name__ == '__main__':

    parameters = {
      'P': 6000.,
      'L': 14.,
      'E': 30.e+6,
      'G': 12.e+6,
      't_max': 13600.,
      's_max': 30000.,
      'd_max': 0.25,
    }

    from mystic.solvers import diffev2
    from mystic.math import almostEqual

    penalty = generate_penalty(**parameters)

    result = diffev2(objective, x0=bounds, bounds=bounds, penalty=penalty, npop=40, gtol=500, disp=False, full_output=True)

    assert almostEqual(result[0], xs, rel=1e-2)
    assert almostEqual(result[1], ys, rel=1e-2)



# EOF



#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Author: Patrick Hung (patrickh @caltech)
# Copyright (c) 1997-2016 California Institute of Technology.
# Copyright (c) 2016-2026 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/mystic/blob/master/LICENSE
__doc__ = _doc = """
This is drawn from Mathematica's example suites, with the 'fOsc3D'
function definition found in [1], and the 'XXX' function found in [2].

References:
    1. Trott, M. "The Mathematica GuideBook for Numerics", Springer-Verlag,
       New York, 2006.
    2. Champion, B. and Strzebonski, A. "Wolfram Mathematica Tutorial
       Collection on Constrained Optimization", Wolfram Research, USA, 2008.
       http://reference.wolfram.com/language/guide/Optimization.html
"""
__all__ = ['fOsc3D', 'NMinimize51', 'fosc3d', 'nmin51']

from .abstract_model import AbstractFunction

from math import sin, exp

class fOsc3D(AbstractFunction):
    __doc__ = \
    """a fOsc3D function generator

A fOsc3D function [1] for positive x_1 values yields
small sinusoidal oscillations on a flat plane, where
a sinkhole containing the global minimum and a few local
minima is found in a small region near the origin.
For negative x_1 values, a parabolic penalty is applied
that decreases as the x_1 appoaches zero.

The generated function f(x) is identical to equation (75)
of section 1.10 of [1], and requires len(x) == 2.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates the fOsc3D function for a list of coeffs

f(x) = f_0(x) + p(x)

Where:
f_0(x) = -4 * \exp(-x_(0)^2 - x_(1)^2) + \sin(6*x_(0)) * \sin(5*x_(1))
with for x_1 < 0:
p(x) = 100.*x_(1)^2
and otherwise:
p(x) = 0.

Inspect with mystic_model_plotter using::
    mystic.models.fosc3d -b "-5:5:.1, 0:5:.1" -d

The minimum is f(x)=-4.501069742528923 at x=(-0.215018, 0.240356)"""
        x,y = coeffs
        func =  -4. * exp( -x*x - y*y ) + sin(6. * x) * sin(5. *y)
        penalty = 0
        if y < 0: penalty = 100.*y*y
        return func + penalty

    minimizers = [(-0.215018, 0.240356)] #XXX: there are many local minima
    pass


class NMinimize51(AbstractFunction):
    __doc__ = \
    """a NMinimize51 function generator

A NMinimize51 function [2] has many local minima. The
minima are periodic over parameter space, and modulate
the surface of a parabola at the coarse scale.  The
global minimum is located at the deepest of the many
periodic wells.

The generated function f(x) is identical to equation (51)
of the 'NMinimize' section in [2], and requires len(x) == 2.
    """ + _doc
    def __init__(self, ndim=2): # is 2-dimensional
        AbstractFunction.__init__(self, ndim=ndim)
        return

    def function(self,coeffs):
        r"""evaluates the NMinimize51 function for a list of coeffs

f(x) = f_0(x) + f_1(x)

Where:
f_0(x) = \exp(\sin(50*x_0)) + \sin(60*\exp(x_1)) + \sin(70*\sin(x_0))
and
f_1(x) = \sin(\sin(80*x_1)) - \sin(10*(x_0 + x_1)) + (x_(0)^2 + x_(1)^2)/4

Inspect with mystic_model_plotter using::
    mystic.models.nmin51 -b "-5:5:.1, 0:5:.1" -d

The minimum is f(x)=-3.306869 at x=(-0.02440313,0.21061247)"""
        x,y = coeffs
        return exp(sin(50.*x)) + sin(60.*exp(y)) + sin(70.*sin(x)) + \
               sin(sin(80.*y)) - sin(10.*(x + y)) + 1./4.*(x**2 + y**2)

    minimizers = [(-0.02440313,0.21061247)] #XXX: there are many local minima
    pass


# cleanup
del _doc

# prepared instances
fosc3d = fOsc3D().function
nmin51 = NMinimize51().function

# End of file
